@@ -1,6 +1,6 @@
 //! An animation system for transitioning between various kinds of values over time.
 
-use crate::cx::*;
+use crate::*;
 use std::f64::consts::PI;
 
 /// Manages animations. Assumes that you always pass it with [`Anim`] objects with
@@ -14,13 +14,14 @@ use std::f64::consts::PI;
 /// Also note that the [`Animator`] always contains the "source of truth" for the
 /// values it manages, so whenever necessary you should copy the values kept
 /// here to the actual "draw objects".
-#[derive(Clone, Debug)]
+#[derive(Debug, Default)]
 pub struct Animator {
-    /// The current source of truth of values that this [`Animator`] manages. Will
-    /// always be set to something, so you can always refer to these. Call
-    /// [`Animator::process_animator`] or [`Animator::handle_animator`] to update these values based on
-    /// the [`Animator::current`] animation and the current time.
-    values: Vec<AnimValue>,
+    /// The current source of truth of values that this [`Animator`] manages.
+    /// Initially undefined, until [`Animator::draw`] initializes the Animator.
+    ///
+    /// Call [`Animator::draw`] or [`Animator::handle`] to update these
+    /// values based on the [`Animator::current`] animation and the current time.
+    values: Option<Vec<AnimValue>>,
 
     /// The current [`Anim`] that is being played. Is [`None`] when there is no
     /// active animation.
@@ -34,31 +35,12 @@ pub struct Animator {
     /// The time that the [`Animator::current`] animation started playing.
     current_start_time: f64,
 
-    /// The values that the previous animation ended with. Used for animations
-    /// that don't have explicit start values; they will start from the values
-    /// where the previous animation ended.
-    prev_anim_end_values: Vec<AnimValue>,
-
     /// The last timestamp we updated our animation, used to avoid computing the
     /// animation values multiple times for the same timestamp.
     last_processed_time: f64,
 }
 
 impl Animator {
-    /// Create a new animator, with the values initialized to the "end state" of
-    /// the `default_anim` that is passed in.
-    pub fn new(default_anim: Anim) -> Self {
-        let def_anim_last_values = default_anim.get_last_values();
-        Self {
-            current: None,
-            next: None,
-            current_start_time: 0.,
-            prev_anim_end_values: def_anim_last_values.clone(),
-            values: def_anim_last_values,
-            last_processed_time: f64::NAN,
-        }
-    }
-
     /// Play an animation. If an animation is already playing, it's either cut
     /// off, or remains playing with the new animation queued up, if [`Anim::chain`]
     /// is set in the new animation. If there was already another animation
@@ -70,9 +52,6 @@ impl Animator {
             self.current = Some(anim);
             self.next = None;
             self.current_start_time = cx.last_event_time;
-            // In case there was a previous animation playing, use its current
-            // `values` as the starting point for the new animation.
-            self.prev_anim_end_values = self.values.clone();
             // Make sure that we request a new frame to play our animation in.
             cx.request_next_frame();
         } else {
@@ -82,13 +61,22 @@ impl Animator {
         }
     }
 
-    /// Convenient function for only calling [`Animator::process_animator`] if the event is
+    /// Process animations from a "draw" function. This must be called before reading any values.
+    ///
+    /// The `anim_default` will initialze the Animator if it's currently uninitialized.
+    pub fn draw(&mut self, cx: &mut Cx, anim_default: Anim) {
+        if self.values.is_none() {
+            self.values = Some(anim_default.get_last_values());
+        }
+        self.run_animator(cx);
+    }
+
+    /// Convenient function for only calling [`Animator::run_animator`] if the event is
     /// an [`Event::NextFrame`]. Returns true if we processed the animation so you
-    /// can update your "draw objects" (or even decide to mark any [`View`]s for
-    /// redraw).
-    pub fn handle_animator(&mut self, cx: &mut Cx, event: &Event) -> bool {
+    /// can update your "draw objects".
+    pub fn handle(&mut self, cx: &mut Cx, event: &Event) -> bool {
         match event {
-            Event::NextFrame => self.process_animator(cx),
+            Event::NextFrame => self.values.is_some() && self.run_animator(cx),
             _ => false,
         }
     }
@@ -97,15 +85,9 @@ impl Animator {
     /// Returns whether or not [`Animator::values`] have been updated, so you can update
     /// your "draw objects" accordingly. Note that [`Animator::values`] are the "source of
     /// truth", so when in doubt it's always safe to just update your objects
-    /// based on [`Animator::values`] regardless of the return value of [`Animator::process_animator`].
-    pub fn process_animator(&mut self, cx: &mut Cx) -> bool {
-        // The very first time that `process_animator` is called we do want to emit
-        // `true`, even though technically nothing has changed since the
-        // initialization. This is just more convenient, so that user code doesn't
-        // have to check if we just initialized the `Animator` or not.
-        let is_initial_call = self.last_processed_time.is_nan();
-
-        // Skip if time hasn't changed.
+    /// based on [`Animator::values`] regardless of the return value of [`Animator::draw`].
+    fn run_animator(&mut self, cx: &mut Cx) -> bool {
+        // Skip if time hasn't changed, unless this is the initial call.
         if cx.last_event_time == self.last_processed_time {
             return false;
         }
@@ -118,8 +100,7 @@ impl Animator {
                 // Make sure that `values` actually reflects the "end state" of the animation, since
                 // at the previous rendering step we were probably a little bit before the actual
                 // end.
-                self.values = current_anim.get_last_values();
-                self.prev_anim_end_values = self.values.clone();
+                self.values = Some(current_anim.get_last_values());
 
                 if self.next.is_none() {
                     // If there was no animation queued up, just bail out, but still return `true`
@@ -147,79 +128,76 @@ impl Animator {
             // Compute the fraction between 0 and 1 of how far we are into the current animation.
             let time_fraction = (cx.last_event_time - self.current_start_time) / current_anim.duration;
 
+            let values = self.values.as_mut().unwrap();
+
             // Update all the individual values based on how far we are into the current animation.
-            // Note that this is also where we look at `prev_anim_end_values`, for if the current
-            // animation doesn't have an explicit starting value.
             for (index, track) in current_anim.tracks.iter().enumerate() {
                 match track {
                     Track::Float { key_frames, ease } => {
-                        self.values[index] = AnimValue::Float(Track::compute_track_float(
+                        values[index] = AnimValue::Float(Track::compute_track_float(
                             time_fraction,
-                            &key_frames,
-                            self.prev_anim_end_values[index].unwrap_float(),
-                            &ease,
+                            key_frames,
+                            values[index].unwrap_float(),
+                            ease,
                         ));
                     }
                     Track::Vec2 { key_frames, ease } => {
-                        self.values[index] = AnimValue::Vec2(Track::compute_track_vec2(
+                        values[index] = AnimValue::Vec2(Track::compute_track_vec2(
                             time_fraction,
-                            &key_frames,
-                            self.prev_anim_end_values[index].unwrap_vec2(),
-                            &ease,
+                            key_frames,
+                            values[index].unwrap_vec2(),
+                            ease,
                         ));
                     }
                     Track::Vec3 { key_frames, ease } => {
-                        self.values[index] = AnimValue::Vec3(Track::compute_track_vec3(
+                        values[index] = AnimValue::Vec3(Track::compute_track_vec3(
                             time_fraction,
-                            &key_frames,
-                            self.prev_anim_end_values[index].unwrap_vec3(),
-                            &ease,
+                            key_frames,
+                            values[index].unwrap_vec3(),
+                            ease,
                         ));
                     }
                     Track::Vec4 { key_frames, ease } => {
-                        self.values[index] = AnimValue::Vec4(Track::compute_track_vec4(
+                        values[index] = AnimValue::Vec4(Track::compute_track_vec4(
                             time_fraction,
-                            &key_frames,
-                            self.prev_anim_end_values[index].unwrap_vec4(),
-                            &ease,
+                            key_frames,
+                            values[index].unwrap_vec4(),
+                            ease,
                         ));
                     }
                 }
             }
             return true;
-        } else {
-            // If there isn't a current animation, then we didn't update `values`. But if this is
-            // the very first call, still return `true` so we'll update the initial values.
-            return is_initial_call;
         }
+        false
     }
 
     /// Get the value of the given track as a float. Be sure to call this only if the given track is
     /// indeed always a float in the [`Anim`]s you pass into this [`Animator`].
     /// TODO(JP): Instead of having multiple functions here, perhaps we can use [`Into`?
     pub fn get_float(&self, track_index: usize) -> f32 {
-        self.values[track_index].unwrap_float()
+        self.values.as_ref().unwrap()[track_index].unwrap_float()
     }
 
     /// Get the value of the given track as a [`Vec2`]. Be sure to call this only if the given track is
     /// indeed always a [`Vec2`] in the [`Anim`]s you pass into this [`Animator`].
     /// TODO(JP): Instead of having multiple functions here, perhaps we can use [`Into`?
     pub fn get_vec2(&self, track_index: usize) -> Vec2 {
-        self.values[track_index].unwrap_vec2()
+        self.values.as_ref().unwrap()[track_index].unwrap_vec2()
     }
 
     /// Get the value of the given track as a [`Vec3`]. Be sure to call this only if the given track is
     /// indeed always a [`Vec3`] in the [`Anim`]s you pass into this [`Animator`].
     /// TODO(JP): Instead of having multiple functions here, perhaps we can use [`Into`?
     pub fn get_vec3(&self, track_index: usize) -> Vec3 {
-        self.values[track_index].unwrap_vec3()
+        self.values.as_ref().unwrap()[track_index].unwrap_vec3()
     }
 
     /// Get the value of the given track as a [`Vec4`]. Be sure to call this only if the given track is
     /// indeed always a [`Vec4`] in the [`Anim`]s you pass into this [`Animator`].
     /// TODO(JP): Instead of having multiple functions here, perhaps we can use [`Into`?
     pub fn get_vec4(&self, track_index: usize) -> Vec4 {
-        self.values[track_index].unwrap_vec4()
+        self.values.as_ref().unwrap()[track_index].unwrap_vec4()
     }
 }
 
@@ -294,7 +272,7 @@ pub struct Anim {
 
 impl Anim {
     /// TODO(JP): Replace these with Anim::default() when
-    /// <https://github.com/rust-lang/rust/pull/86808> gets done
+    /// <https://github.com/rust-lang/rust/issues/67792> gets done
     pub const DEFAULT: Anim = Anim { duration: 0., chain: false, tracks: &[] };
 
     /// Get the values for the "end state" of an animation, ie. the values for
@@ -359,7 +337,7 @@ pub enum Ease {
 }
 impl Ease {
     /// TODO(JP): Replace these with Ease::default() when
-    /// <https://github.com/rust-lang/rust/pull/86808> gets done
+    /// <https://github.com/rust-lang/rust/issues/67792> gets done
     pub const DEFAULT: Ease = Ease::InOutCubic;
 }
 impl Default for Ease {
@@ -369,6 +347,8 @@ impl Default for Ease {
 }
 
 impl Ease {
+    // Clippy TODO
+    #[warn(clippy::many_single_char_names)]
     fn map(&self, t: f64) -> f64 {
         match self {
             Ease::Lin => t.max(0.0).min(1.0),
