@@ -1,75 +1,19 @@
 //! The main module of this crate. Contains [`Cx`], which you will see
 //! used all over the place.
-//!
-//! This re-exports a lot of stuff, and so internally you
-//! will often see us write `use crate::cx::*` (external users should
-//! instead write `use zaplib::*`, which will include all of
-//! the exports in here as well).
 
-use std::collections::{BTreeSet, HashMap};
-use std::fmt::Write;
+use crate::*;
+
+#[cfg(feature = "cef")]
+use cef_browser::MaybeCefBrowser;
+use debug_log::DebugLog;
+use std::{
+    any::{Any, TypeId},
+    collections::{BTreeSet, HashMap},
+    fmt::Write,
+    sync::{Arc, RwLock},
+};
 use zaplib_shader_compiler::generate_shader_ast::*;
 
-pub use zaplib_shader_compiler::generate_shader_ast::CodeFragment;
-pub use zaplib_shader_compiler::math::*;
-pub use zaplib_shader_compiler::ty::Ty;
-
-pub use crate::animator::*;
-pub use crate::area::*;
-pub use crate::cursor::*;
-pub use crate::draw_input_type::*;
-pub use crate::draw_tree::*;
-pub use crate::events::*;
-pub use crate::fonts::*;
-pub use crate::geometry::*;
-pub use crate::hash::*;
-pub use crate::macros::*;
-pub use crate::menu::*;
-pub use crate::pass::*;
-pub use crate::read_seek::*;
-pub use crate::shader::*;
-pub use crate::texture::*;
-pub use crate::turtle::*;
-pub use crate::universal_instant::*;
-pub use crate::universal_thread::*;
-pub use crate::window::*;
-
-#[cfg(all(not(feature = "ipc"), target_os = "linux"))]
-pub(crate) use crate::cx_linux::*;
-#[cfg(all(not(feature = "ipc"), target_os = "linux"))]
-pub(crate) use crate::cx_opengl::*;
-
-#[cfg(all(not(feature = "ipc"), target_os = "macos"))]
-pub(crate) use crate::cx_macos::*;
-#[cfg(all(not(feature = "ipc"), target_os = "macos"))]
-pub(crate) use crate::cx_metal::*;
-
-#[cfg(all(not(feature = "ipc"), target_os = "windows"))]
-pub(crate) use crate::cx_dx11::*;
-#[cfg(all(not(feature = "ipc"), target_os = "windows"))]
-pub(crate) use crate::cx_windows::*;
-
-#[cfg(all(not(feature = "ipc"), target_arch = "wasm32"))]
-pub(crate) use crate::cx_webgl::*;
-
-#[cfg(all(not(feature = "ipc"), any(target_os = "linux", target_os = "macos", target_os = "windows")))]
-pub(crate) use crate::cx_desktop::*;
-
-#[cfg(all(not(feature = "ipc"), target_arch = "wasm32"))]
-pub use crate::cx_wasm32::*;
-
-#[cfg(feature = "ipc")]
-pub use crate::cx_ipc_child::*;
-
-#[cfg(all(feature = "ipc", target_arch = "wasm32"))]
-pub use crate::cx_ipc_wasm32::*;
-
-#[cfg(all(feature = "ipc", any(target_os = "linux", target_os = "macos")))]
-pub use crate::cx_ipc_posix::*;
-
-#[cfg(all(feature = "ipc", target_os = "windows"))]
-pub use crate::cx_ipc_win32::*;
-
 /// Contains information about the platform (operating system) that we're running on.
 #[derive(Clone)]
 pub enum PlatformType {
@@ -92,6 +36,9 @@ impl PlatformType {
     }
 }
 
+#[cfg(any(feature = "cef", target_arch = "wasm32"))]
+pub type CallRustInSameThreadSyncFn = fn(name: &str, params: Vec<ZapParam>) -> Vec<ZapParam>;
+
 /// The main "context" object which contains pretty much everything we need within the framework.
 pub struct Cx {
     /// See [`PlatformType`].
@@ -118,16 +65,16 @@ pub struct Cx {
     /// TODO(JP): We currently never remove old [`CxView`]s.
     pub(crate) views: Vec<CxView>,
 
-    /// The compiled [`CxShader`]s. [`Shader::shader_id`] represents an index in this list.
+    /// The compiled [`CxShader`]s. [`Shader::shader_id`] and [`Shader::shader_id`] represent an index in this list.
     pub(crate) shaders: Vec<CxShader>,
-    /// Map from hashed names (using [`StringHash`] to shader_id / indices in [`Cx::shaders`].
-    pub(crate) shader_id_by_hashed_name: HashMap<u64, usize>,
-    /// List of actual [`CxTexture`] objects. [`Texture::texture_id`] represents an index in this list.
+    /// Shader IDs (indices in [`Cx::shaders`]) that need to be recompiled.
+    pub(crate) shader_recompile_ids: Vec<usize>,
+    /// List of actual [`CxTexture`] objects. [`TextureHandle::texture_id`] represents an index in this list.
     pub(crate) textures: Vec<CxTexture>,
-    /// List of actual [`CxGeometry`] objects. [`Geometry::geometry_id`] represents an index in this list.
-    pub(crate) geometries: Vec<CxGeometry>,
+    /// List of actual [`CxGpuGeometry`] objects. [`GpuGeometry::gpu_geometry_id`] represents an index in this list.
+    pub(crate) gpu_geometries: Vec<CxGpuGeometry>,
 
-    /// Whether we are currently (re)drawing, ie. we called the app's `draw_app` function.
+    /// Whether we are currently (re)drawing, ie. we called the app's `draw` function.
     pub(crate) in_redraw_cycle: bool,
     /// An auto-incrementing ID representing the current (re)draw cycle.
     ///
@@ -143,32 +90,24 @@ pub struct Cx {
     /// Stack of [`View::view_id`]s / indices into [`Cx::views`], using [`View::begin_view`]
     /// and [`View::end_view`].
     pub(crate) view_stack: Vec<usize>,
-    /// A stack of [`Turtle`]s, using [`Cx::begin_turtle`] and [`Cx::end_turtle`]. Turtles all the
-    /// way down!
-    pub(crate) turtles: Vec<Turtle>,
+    /// A stack of [`CxLayoutBox`]s, using [`Cx::begin_typed_box`] and [`Cx::end_typed_box`]
+    pub(crate) layout_boxes: Vec<CxLayoutBox>,
+
+    /// The instance offsets for the different [`Shader`]s when the current "shader group" was started.
+    ///
+    /// Empty when there is no current "shader group". See [`Cx::begin_shader_group`].
+    pub(crate) shader_group_instance_offsets: Vec<usize>,
 
     /// A list of [`Area`]s that we want to align later on. This is kept separate
-    /// from [`Cx::turtles`] (even though this is part of the turtle layout
-    /// system) so that a parent [`Turtle`] can align a bunch of stuff that was
-    /// drawn using child [`Turtle`]s.
+    /// from [`Cx::layout_boxes`] (even though this is part of the box layout
+    /// system) so that a parent [`CxLayoutBox`] can align a bunch of stuff that was
+    /// drawn using child [`CxLayoutBox`]s.
     ///
     /// TODO(JP): This may currently only contain [`Area::InstanceRange`], so
     /// maybe we should change the type of this? It might also be nice to be able
     /// explicitly push [`Area::View`] on this list though, and then set a uniform
     /// on the entire [`CxView`], for better performance?
-    pub(crate) turtle_align_list: Vec<Area>,
-
-    /// List of [`Area`]s that need to be redrawn on the next draw cycle.
-    /// See [`Cx::redraw_child_area`].
-    pub(crate) redraw_child_areas: Vec<Area>,
-    /// List of [`Area`]s that need to be redrawn in the current draw cycle.
-    /// This is kept separate from [`Cx::redraw_child_areas`] so that we can mark
-    /// new [`Area`]s for redraw _during_ a draw cycle.
-    ///
-    /// TODO(JP): See [`Cx::redraw_child_area`] for a TODO involving this field.
-    ///
-    /// See [`Cx::redraw_child_area`].
-    _redraw_child_areas: Vec<Area>,
+    pub(crate) layout_box_align_list: Vec<Area>,
 
     /// The system-default `dpi_factor`. See also [`PassUniforms::dpi_factor`].
     ///
@@ -179,9 +118,9 @@ pub struct Cx {
     /// See also [`PassUniforms::dpi_factor`].
     ///
     /// More commonly known as the "device pixel ratio". TODO(JP): Rename?
-    pub(crate) current_dpi_factor: f32,
+    pub current_dpi_factor: f32,
 
-    /// Last timestamp from when an event was fired, in second since the application
+    /// Last timestamp from when an event was fired, in seconds since the application
     /// was started. Typically you want to use this instead of making a system call.
     pub last_event_time: f64,
 
@@ -190,16 +129,16 @@ pub struct Cx {
     /// The last [`Signal::signal_id`] that was issued.
     pub(crate) last_signal_id: usize,
 
-    /// The current [`Area`] that has keyboard focus, so it can register key input [`Event`]s.
+    /// The current [`ComponentId`] that has keyboard focus, so it can register key input [`Event`]s.
     ///
     /// See also [`Cx::prev_key_focus`] and [`Cx::next_key_focus`].
-    pub(crate) key_focus: Area,
-    /// The [`Area`] that previously was [`Cx::key_focus`], so you can revert it using
+    pub(crate) key_focus: Option<ComponentId>,
+    /// The [`ComponentId`] that previously was [`Cx::key_focus`], so you can revert it using
     /// [`Cx::revert_key_focus`].
     ///
     /// See also [`Cx::key_focus`] and [`Cx::next_key_focus`].
-    prev_key_focus: Area,
-    /// The [`Area`] that will become [`Cx::key_focus`] when the current events are handled.
+    prev_key_focus: Option<ComponentId>,
+    /// The [`ComponentId`] that will become [`Cx::key_focus`] when the current events are handled.
     /// Gets set using [`Cx::set_key_focus`] or [`Cx::revert_key_focus`].
     ///
     /// See also [`Cx::prev_key_focus`] and [`Cx::next_key_focus`].
@@ -207,26 +146,29 @@ pub struct Cx {
     /// TODO(JP): It's possible to set this during the draw cycle instead of during an
     /// event handler, and then it won't update [`Cx::key_focus`] until the next event
     /// is handled. We should probably guard against that.
-    next_key_focus: Area,
+    next_key_focus: Option<Option<ComponentId>>,
     pub(crate) keys_down: Vec<KeyEvent>,
 
     /// The cursor type that the user sees while holding the mouse down. Gets reset to [`None`] when
-    /// you release the mouse button ([`Event::FingerUp`]).
+    /// you release the mouse button ([`Event::PointerUp`]).
     pub(crate) down_mouse_cursor: Option<MouseCursor>,
 
     /// The cursor type that the user sees while hovering (not holding the mouse down).
-    /// Gets reset when there's a new [`Event::FingerHover`], so you have to periodically set this.
+    /// Gets reset when there's a new [`Event::PointerHover`], so you have to periodically set this.
     pub(crate) hover_mouse_cursor: Option<MouseCursor>,
 
-    /// The current state of each "finger" that we track.
+    /// The current state of each "pointer" that we track.
     ///
     /// TODO(JP): This seems mostly relevant for multi-touch, which we don't really support very
     /// well yet. Should we keep this?
-    pub(crate) fingers: Vec<CxPerFinger>,
+    pub(crate) pointers: Vec<CxPerPointer>,
 
     /// Whether [`Cx::request_next_frame`] was called.
     pub(crate) requested_next_frame: bool,
 
+    /// Whether [`Cx::request_draw`] was called.
+    pub(crate) requested_draw: bool,
+
     /// The local "signals", which are like custom events that also work across threads.
     ///
     /// See also [`Signal`] and [`SignalEvent`].
@@ -253,10 +195,67 @@ pub struct Cx {
     /// The user's event handler. Storing it like this cuts the compile time of an end-user application in half.
     pub(crate) event_handler: Option<*mut dyn FnMut(&mut Cx, &mut Event)>,
 
-    /// List of actual [`CxFont`] objects. [`Font::font_id`] represents an index in this list.
-    pub(crate) fonts: Vec<CxFont>,
-    /// See [`CxFontsAtlas`].
-    pub(crate) fonts_atlas: CxFontsAtlas,
+    /// Fonts specific data
+    /// It might be possible for fonts data to be shared between different threads, so
+    /// we need to make use of locks.
+    pub fonts_data: Arc<RwLock<CxFontsData>>,
+
+    /// A buffer with temporary data used in [`Area::get_first`] and [`Area::get_first_mut`].
+    ///
+    /// Shouldn't be used excessively. Gets cleared out after a draw cycle.
+    ///
+    /// TODO(JP): It would be nice if we can eliminate this altogether, e.g. by guaranteeing
+    /// that handle functions are only called after a component has been drawn, or by making
+    /// shader animations a more integral part of the framework.
+    pub(crate) temp_default_data: Vec<Box<dyn Any>>,
+
+    /// See [`CxDebugFlags`].
+    pub(crate) debug_flags: CxDebugFlags,
+
+    #[cfg(feature = "cef")]
+    pub(crate) cef_browser: MaybeCefBrowser,
+
+    /// Various debug logs that are getting appended during draw cycle.
+    /// See [`DebugLog`] for more information on supported types
+    pub(crate) debug_logs: Vec<DebugLog>,
+
+    /// Function registered through [`Cx::on_call_rust`]
+    pub call_rust_fn: Option<usize>,
+
+    /// Reference to the main_app type
+    pub app_type_id: TypeId,
+}
+
+/// Flags that can be set that enable debug functionality. See [`Cx::debug_flags_mut`] for an example.
+#[derive(Copy, Clone, Default)]
+pub struct CxDebugFlags {
+    /// See [`CxDebugDrawTree`].
+    pub draw_tree: CxDebugDrawTree,
+
+    /// Makes it so every call to `Cx::add_instances` gets a fresh [`DrawCall`]. This is useful for debugging,
+    /// since the batching of draw calls can be confusing sometimes (and you should never rely on it happening).
+    pub disable_draw_call_batching: bool,
+
+    /// Enables overlay with borders of CxLayoutBox rects
+    pub enable_layout_debugger: bool,
+}
+
+/// What kind of debug information should be printed about the draw tree.
+#[derive(Copy, Clone, PartialEq, Debug)]
+pub enum CxDebugDrawTree {
+    /// No draw tree debugging.
+    None,
+    /// Print the draw tree.
+    DrawTree,
+    /// Print the draw tree and also information on the individual instances.
+    Instances,
+    /// Print geometries.
+    Geometries,
+}
+impl Default for CxDebugDrawTree {
+    fn default() -> Self {
+        Self::None
+    }
 }
 
 /// Settings for "commands"; see [`CommandId`].
@@ -271,76 +270,75 @@ pub(crate) struct CxCommandSetting {
 }
 
 #[derive(Default, Clone)]
-pub(crate) struct CxPerFinger {
-    pub(crate) captured: Area,
+pub(crate) struct CxPerPointer {
+    pub(crate) captured: Option<ComponentId>,
     pub(crate) tap_count: (Vec2, f64, u32),
     pub(crate) down_abs_start: Vec2,
     pub(crate) down_rel_start: Vec2,
-    pub(crate) over_last: Area,
-    pub(crate) _over_last: Area,
+    pub(crate) over_last: Option<ComponentId>,
+    pub(crate) _over_last: Option<ComponentId>,
 }
 
-pub(crate) const NUM_FINGERS: usize = 10;
+pub(crate) const NUM_POINTERS: usize = 10;
 
-impl Default for Cx {
-    fn default() -> Self {
-        let mut fingers = Vec::new();
-        fingers.resize(NUM_FINGERS, CxPerFinger::default());
+impl Cx {
+    pub fn new(app_type_id: TypeId) -> Self {
+        let mut pointers = Vec::new();
+        pointers.resize(NUM_POINTERS, CxPerPointer::default());
 
         let textures = vec![CxTexture {
-            desc: TextureDesc { format: TextureFormat::ImageBGRA, width: Some(4), height: Some(4), multisample: None },
+            desc: TextureDesc { format: TextureFormat::ImageRGBA, width: Some(4), height: Some(4), multisample: None },
             image_u32: vec![0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
             update_image: true,
             platform: CxPlatformTexture::default(),
         }];
 
+        // We try to initialize Vecs with some reasonable capacity, to prevent reallocations.
         Self {
             platform_type: PlatformType::Unknown,
 
             windows: Vec::new(),
             windows_free: Vec::new(),
-            passes: Vec::new(),
+            passes: Vec::with_capacity(10),
             // TODO(JP): See my note up at [`Cx::views`].
-            views: vec![CxView { ..Default::default() }],
-            fonts: Vec::new(),
-            fonts_atlas: CxFontsAtlas::default(),
+            views: vec![CxView::default()],
+            fonts_data: Arc::new(RwLock::new(CxFontsData::default())),
             textures,
-            shaders: Vec::new(),
-            shader_id_by_hashed_name: HashMap::new(),
-            geometries: Vec::new(),
+            shaders: Vec::with_capacity(50),
+            shader_recompile_ids: Vec::with_capacity(50),
+            gpu_geometries: Vec::new(),
 
             default_dpi_factor: 1.0,
             current_dpi_factor: 1.0,
             in_redraw_cycle: false,
             window_stack: Vec::new(),
-            pass_stack: Vec::new(),
-            view_stack: Vec::new(),
-            turtles: Vec::new(),
-            turtle_align_list: Vec::new(),
+            pass_stack: Vec::with_capacity(10),
+            view_stack: Vec::with_capacity(50),
+            layout_boxes: Vec::with_capacity(100),
+            layout_box_align_list: Vec::with_capacity(100),
+            shader_group_instance_offsets: Vec::with_capacity(10),
 
             last_event_time: 0.0,
 
-            redraw_child_areas: Vec::new(),
-            _redraw_child_areas: Vec::new(),
-
             redraw_id: 1,
             last_timer_id: 1,
             last_signal_id: 1,
 
-            next_key_focus: Area::Empty,
-            prev_key_focus: Area::Empty,
-            key_focus: Area::Empty,
+            next_key_focus: None,
+            prev_key_focus: None,
+            key_focus: None,
             keys_down: Vec::new(),
 
             down_mouse_cursor: None,
             hover_mouse_cursor: None,
-            fingers,
+            pointers,
 
             shader_ast_generator: ShaderAstGenerator::new(),
 
             command_settings: HashMap::new(),
 
             requested_next_frame: false,
+            requested_draw: false,
 
             profiles: HashMap::new(),
 
@@ -348,25 +346,116 @@ impl Default for Cx {
 
             panic_redraw: false,
 
-            platform: CxPlatform { ..Default::default() },
+            platform: CxPlatform::default(),
 
             event_handler: None,
+
+            temp_default_data: Vec::with_capacity(1000),
+
+            debug_flags: Default::default(),
+
+            #[cfg(feature = "cef")]
+            cef_browser: MaybeCefBrowser::new(),
+
+            debug_logs: Vec::new(),
+
+            call_rust_fn: None,
+            app_type_id,
+        }
+    }
+
+    pub(crate) fn process_pre_event(&mut self, event: &mut Event) {
+        match event {
+            Event::PointerHover(pe) => {
+                self.pointers[pe.digit].over_last = None;
+                self.hover_mouse_cursor = None;
+            }
+            Event::PointerUp(_pe) => {
+                self.down_mouse_cursor = None;
+            }
+            Event::PointerDown(pe) => {
+                // lets set the pointer tap count
+                pe.tap_count = self.process_tap_count(pe.digit, pe.abs, pe.time);
+            }
+            Event::KeyDown(ke) => {
+                self.process_key_down(ke.clone());
+                if ke.key_code == KeyCode::PrintScreen {
+                    self.panic_redraw = true;
+                }
+
+                // shortcuts: ctrl + option + cmd + ...
+                if ke.modifiers.control && ke.modifiers.alt && ke.modifiers.logo {
+                    match ke.key_code {
+                        KeyCode::Key1 => {
+                            self.debug_flags.enable_layout_debugger = !self.debug_flags.enable_layout_debugger;
+                            log!("Set enable_layout_debugger to {}", self.debug_flags.enable_layout_debugger);
+                            self.request_draw();
+                        }
+                        KeyCode::Key2 => {
+                            self.debug_flags.disable_draw_call_batching = !self.debug_flags.disable_draw_call_batching;
+                            log!("Set disable_draw_call_batching to {}", self.debug_flags.disable_draw_call_batching);
+                            self.request_draw();
+                        }
+                        KeyCode::Key3 => {
+                            // cycle through options:
+                            match self.debug_flags.draw_tree {
+                                CxDebugDrawTree::None => self.debug_flags.draw_tree = CxDebugDrawTree::DrawTree,
+                                CxDebugDrawTree::DrawTree => self.debug_flags.draw_tree = CxDebugDrawTree::Instances,
+                                CxDebugDrawTree::Instances => self.debug_flags.draw_tree = CxDebugDrawTree::Geometries,
+                                CxDebugDrawTree::Geometries => self.debug_flags.draw_tree = CxDebugDrawTree::None,
+                            }
+                            log!("Set draw_tree to {:?}", self.debug_flags.draw_tree);
+                            self.request_draw();
+                        }
+                        _ => {}
+                    }
+                }
+            }
+            Event::KeyUp(ke) => {
+                self.process_key_up(ke);
+            }
+            Event::AppFocusLost => {
+                self.call_all_keys_up();
+            }
+            _ => (),
+        };
+    }
+
+    pub(crate) fn process_post_event(&mut self, event: &mut Event) {
+        match event {
+            Event::PointerUp(pe) => {
+                // decapture automatically
+                self.pointers[pe.digit].captured = None;
+            }
+            Event::PointerHover(pe) => {
+                // new last area pointer over
+                self.pointers[pe.digit]._over_last = self.pointers[pe.digit].over_last;
+                //if pe.hover_state == HoverState::Out{
+                //    self.hover_mouse_cursor = None;
+                //}
+            }
+            Event::PointerScroll(_) => {
+                // check for anything being paint or draw dirty
+                #[cfg(not(target_arch = "wasm32"))]
+                if self.requested_draw {
+                    self.platform.desktop.repaint_via_scroll_event = true;
+                }
+            }
+            _ => {}
         }
     }
-}
 
-impl Cx {
     pub(crate) fn process_tap_count(&mut self, digit: usize, pos: Vec2, time: f64) -> u32 {
-        if digit >= self.fingers.len() {
+        if digit >= self.pointers.len() {
             return 0;
         };
-        let (last_pos, last_time, count) = self.fingers[digit].tap_count;
+        let (last_pos, last_time, count) = self.pointers[digit].tap_count;
 
         if (time - last_time) < 0.5 && pos.distance(&last_pos) < 10. {
-            self.fingers[digit].tap_count = (pos, time, count + 1);
+            self.pointers[digit].tap_count = (pos, time, count + 1);
             count + 1
         } else {
-            self.fingers[digit].tap_count = (pos, time, 1);
+            self.pointers[digit].tap_count = (pos, time, 1);
             1
         }
     }
@@ -464,135 +553,28 @@ impl Cx {
         }
     }
 
-    /// Mark a particular `Area` to be redrawn.
-    ///
-    /// All parent [`View`]s will also be redrawn, since rendering always starts
-    /// at the root of a [`Window`] or [`Pass`].
-    ///
-    /// If you pass in [`Area::All`], we'll redraw everything. This should
-    /// generally be OK to do, though it is slower than only redrawing
-    /// particular [`Area`]s.
-    ///
-    /// TODO(JP): When called _during_ a draw cycle, this will always trigger a
-    /// new draw cycle, even if the [`Area`] that was marked hasn't been drawn
-    /// yet. It would be better to see if we can still redraw the [`Area`] (and
-    /// its parents) during the current draw cycle, and then modify
-    /// [`Cx::_redraw_child_areas`] instead of [`Cx::redraw_child_areas`].
-    ///
-    /// See also [`View::redraw_view`], [`Cx::view_will_redraw`].
-    pub fn redraw_child_area(&mut self, area: Area) {
+    /// Request a new redraw of the application.
+    pub fn request_draw(&mut self) {
         if self.panic_redraw {
             #[cfg(debug_assertions)]
             panic!("Panic Redraw triggered")
         }
-
-        assert_ne!(area, Area::Empty);
-
-        // if we are redrawing all, clear the rest
-        if area == Area::All {
-            self.redraw_child_areas.truncate(0);
-        }
-        // check if we are already redrawing all
-        else if self.redraw_child_areas.len() == 1 && self.redraw_child_areas[0] == Area::All {
-            return;
-        };
-        // only add it if we dont have it already
-        if self.redraw_child_areas.iter().any(|a| *a == area) {
-            return;
-        }
-        self.redraw_child_areas.push(area);
+        self.requested_draw = true;
     }
 
-    /// Checks if a [`View`] will redraw.
-    ///
-    /// By walking the tree upwards from any [`Area`]s that have been marked for
-    /// redraw in [`Cx::redraw_child_area`].
-    ///
-    /// TODO(JP): This will often visit the same draw tree nodes multiple times,
-    /// so it would be better to cache this somewhere, e.g. in a `Vec<bool>` of
-    /// same size as [`Cx::views`]. We could even precompute that in
-    /// [`Cx::redraw_child_area`] and have that replace [`Cx::redraw_child_areas`]
-    /// entirely.
-    ///
-    /// See also [`View::view_will_redraw`].
-    pub(crate) fn view_will_redraw(&self, view_id: usize) -> bool {
-        // figure out if areas are in some way a child of draw_list_id, then we need to redraw
-        for area in &self._redraw_child_areas {
-            match area {
-                Area::All => {
-                    return true;
-                }
-                Area::Empty => (),
-                Area::InstanceRange(instance) => {
-                    let mut vw = instance.view_id;
-                    if vw == view_id {
-                        return true;
-                    }
-                    while vw != 0 {
-                        vw = self.views[vw].nesting_view_id;
-                        if vw == view_id {
-                            return true;
-                        }
-                    }
-                }
-                Area::View(viewarea) => {
-                    let mut vw = viewarea.view_id;
-                    if vw == view_id {
-                        return true;
-                    }
-                    while vw != 0 {
-                        vw = self.views[vw].nesting_view_id;
-                        if vw == view_id {
-                            return true;
-                        }
-                    }
-                }
-            }
-        }
-        false
-    }
-
-    /// Update any global pointers that were referring to an old [`Area`].
-    ///
-    /// TODO(JP): It's easy to forget to call this! Should we add a field on
-    /// [`Area`] (or use an existing one, like [`ViewArea::redraw_id`]) to
-    /// only mark a new [`Area`] as valid once you pass it through this function?
-    pub fn update_area_refs(&mut self, old_area: Area, new_area: Area) -> Area {
-        if old_area == Area::Empty || old_area == Area::All {
-            return new_area;
-        }
-
-        for finger in &mut self.fingers {
-            if finger.captured == old_area {
-                finger.captured = new_area;
-            }
-            if finger._over_last == old_area {
-                finger._over_last = new_area;
-            }
-        }
-        // update capture keyboard
-        if self.key_focus == old_area {
-            self.key_focus = new_area
-        }
-
-        // update capture keyboard
-        if self.prev_key_focus == old_area {
-            self.prev_key_focus = new_area
-        }
-        if self.next_key_focus == old_area {
-            self.next_key_focus = new_area
-        }
-
-        new_area
-    }
-
-    /// Sets an [`Area`] that will become [`Cx::key_focus`] when the current events are handled.
+    /// Sets a [`ComponentId`] that will become [`Cx::key_focus`] when the current events are handled.
     ///
     /// TODO(JP): It's possible to set this during the draw cycle instead of during an
     /// event handler, and then it won't update [`Cx::key_focus`] until the next event
     /// is handled. We should probably guard against that.
-    pub fn set_key_focus(&mut self, focus_area: Area) {
-        self.next_key_focus = focus_area;
+    pub fn set_key_focus(&mut self, focus_component_id: Option<ComponentId>) {
+        self.next_key_focus = Some(focus_component_id);
+    }
+
+    /// Keep the existing key focus during an [`Event::PointerDown`], because otherwise we'll reset
+    /// it back to [`Area::Empty`].
+    pub fn keep_key_focus(&mut self) {
+        self.next_key_focus = Some(self.key_focus);
     }
 
     /// Reverts back to the previous [`Cx::key_focus`] value.
@@ -601,12 +583,12 @@ impl Cx {
     /// event handler, and then it won't update [`Cx::key_focus`] until the next event
     /// is handled. We should probably guard against that.
     pub fn revert_key_focus(&mut self) {
-        self.next_key_focus = self.prev_key_focus;
+        self.next_key_focus = Some(self.prev_key_focus);
     }
 
-    /// Check if an [`Area`] currently is the [Cx::key_focus] area.
-    pub fn has_key_focus(&self, focus_area: Area) -> bool {
-        self.key_focus == focus_area
+    /// Check if a [`ComponentId`] currently has keyboard focus.
+    pub fn has_key_focus(&self, component_id: Option<ComponentId>) -> bool {
+        self.key_focus == component_id
     }
 
     pub(crate) fn process_key_down(&mut self, key_event: KeyEvent) {
@@ -640,23 +622,50 @@ impl Cx {
             (*event_handler)(self, event);
         }
 
-        if self.next_key_focus != self.key_focus {
-            self.prev_key_focus = self.key_focus;
-            self.key_focus = self.next_key_focus;
-            unsafe {
-                (*event_handler)(self, &mut Event::KeyFocus(KeyFocusEvent { prev: self.prev_key_focus, focus: self.key_focus }));
+        // Someone has to call `set_key_focus` or `keep_key_focus` when handling `PointerDown`, otherwise
+        // the key focus will be reset.
+        if let Event::PointerDown(_) = event {
+            if self.next_key_focus.is_none() {
+                self.next_key_focus = Some(None);
             }
         }
+
+        if let Some(next_key_focus) = self.next_key_focus {
+            if next_key_focus != self.key_focus {
+                self.prev_key_focus = self.key_focus;
+                self.key_focus = next_key_focus;
+                unsafe {
+                    (*event_handler)(
+                        self,
+                        &mut Event::KeyFocus(KeyFocusEvent { prev: self.prev_key_focus, focus: self.key_focus }),
+                    );
+                }
+            }
+            self.next_key_focus = None;
+        }
+
+        self.temp_default_data.clear();
     }
 
     pub(crate) fn call_draw_event(&mut self) {
         // self.profile();
         self.in_redraw_cycle = true;
         self.redraw_id += 1;
-        std::mem::swap(&mut self._redraw_child_areas, &mut self.redraw_child_areas);
-        self.turtle_align_list.truncate(0);
-        self.redraw_child_areas.truncate(0);
-        self.call_event_handler(&mut Event::Draw);
+        self.layout_box_align_list.clear();
+        self.debug_logs.clear();
+
+        // TODO(Paras): Terrible hack.
+        //
+        // For some reason our wasm builds always have an incorrect initial draw, but are immediately
+        // fixed on the second draw call. So here, we make sure we have drawn twice before clearing
+        // `requested_draw`.
+        // Interestingly, native builds do not strictly need this hack, because they draw twice
+        // anyways unlike WASM.
+        if self.redraw_id > 2 {
+            self.requested_draw = false;
+        }
+
+        self.call_event_handler(&mut Event::SystemEvent(SystemEvent::Draw));
         self.in_redraw_cycle = false;
         if !self.view_stack.is_empty() {
             panic!("View stack disaligned, forgot an end_view(cx)");
@@ -667,8 +676,11 @@ impl Cx {
         if !self.window_stack.is_empty() {
             panic!("Window stack disaligned, forgot an end_window(cx)");
         }
-        if !self.turtles.is_empty() {
-            panic!("Turtle stack disaligned, forgot an end_turtle()");
+        if !self.layout_boxes.is_empty() {
+            panic!("LayoutBoxes stack disaligned, forgot an end_box(cx)");
+        }
+        if !self.shader_group_instance_offsets.is_empty() {
+            panic!("Shader group stack disaligned, forgot an end_shader_group()");
         }
         //self.profile();
     }
@@ -730,20 +742,43 @@ impl Cx {
         }
     }
 
-    pub fn status_http_send_ok() -> StatusId {
-        uid!()
-    }
-    pub fn status_http_send_fail() -> StatusId {
-        uid!()
+    pub const STATUS_HTTP_SEND_OK: StatusId = location_hash!();
+    pub const STATUS_HTTP_SEND_FAIL: StatusId = location_hash!();
+
+    /// Change the debug flags, which control various debug functionality.
+    /// See [`CxDebugFlags`] for more information on the individual flags.
+    ///
+    /// Example:
+    /// ```
+    /// let flags = cx.debug_flags_mut();
+    /// flags.draw_tree = CxDebugDrawTree::Instances;
+    /// flags.disable_draw_call_batching = true;
+    /// ```
+    pub fn debug_flags_mut(&mut self) -> &mut CxDebugFlags {
+        &mut self.debug_flags
     }
 
     pub(crate) fn debug_draw_tree(&mut self, view_id: usize) {
-        if let Some(debug_type) = &self.views[view_id].debug {
+        if self.debug_flags.draw_tree == CxDebugDrawTree::DrawTree || self.debug_flags.draw_tree == CxDebugDrawTree::Instances {
             let mut s = String::new();
-            let dump_instances = *debug_type == CxViewDebug::Instances;
+            let dump_instances = self.debug_flags.draw_tree == CxDebugDrawTree::Instances;
             self.debug_draw_tree_recur(dump_instances, &mut s, view_id, 0);
             crate::log!("{}", &s);
         }
+        if self.debug_flags.draw_tree == CxDebugDrawTree::Geometries {
+            crate::log!("--------------- Geometries for redraw_id: {} ---------------", self.redraw_id);
+            for (index, gpu_geometry) in self.gpu_geometries.iter().enumerate() {
+                crate::log!(
+                    "{}: vertex data: {} bytes, index data: {} bytes / {} triangles, dirty: {}, usage_count: {}",
+                    index,
+                    gpu_geometry.geometry.vertices_f32_slice().len() * std::mem::size_of::<f32>(),
+                    gpu_geometry.geometry.indices_u32_slice().len() * std::mem::size_of::<f32>(),
+                    gpu_geometry.geometry.indices_u32_slice().len() / 3,
+                    gpu_geometry.dirty,
+                    gpu_geometry.usage_count()
+                );
+            }
+        }
     }
 
     fn debug_draw_tree_recur(&mut self, dump_instances: bool, s: &mut String, view_id: usize, depth: usize) {
@@ -773,7 +808,7 @@ impl Cx {
             } else {
                 let cxview = &mut self.views[view_id];
                 let draw_call = &mut cxview.draw_calls[draw_call_id];
-                let sh = &self.shaders[draw_call.shader.shader_id];
+                let sh = &self.shaders[draw_call.shader_id];
                 let slots = sh.mapping.instance_props.total_slots;
                 let instances = draw_call.instances.len() / slots;
                 writeln!(
@@ -782,7 +817,7 @@ impl Cx {
                     indent,
                     draw_call_id,
                     sh.name,
-                    draw_call.shader.shader_id,
+                    draw_call.shader_id,
                     instances,
                     vec2(draw_call.draw_uniforms.draw_scroll_x, draw_call.draw_uniforms.draw_scroll_y),
                     vec2(draw_call.draw_uniforms.draw_local_scroll_x, draw_call.draw_uniforms.draw_local_scroll_y)
@@ -829,6 +864,21 @@ impl Cx {
             writeln!(s, "---------- End Debug draw tree for redraw_id: {} ---------", self.redraw_id).unwrap();
         }
     }
+
+    /// Register function to handle callRust from JavaScript. Registered function must be a method on the main app.
+    pub fn on_call_rust<T: 'static>(
+        &mut self,
+        func: fn(this: &mut T, cx: &mut Cx, name: String, params: Vec<ZapParam>) -> Vec<ZapParam>,
+    ) {
+        if self.call_rust_fn.is_some() {
+            panic!("Attempting to call on_call_rust twice.");
+        }
+
+        if self.app_type_id != TypeId::of::<T>() {
+            panic!("Error in on_call_rust: Function must be a method on the main_app.");
+        }
+        self.call_rust_fn = Some(Box::into_raw(Box::new(func)) as usize);
+    }
 }
 
 /// A bunch of traits that are common between the native platforms and the WebAssembly platform. This trait makes sure
@@ -838,20 +888,6 @@ pub trait CxDesktopVsWasmCommon {
     /// TODO(JP): This doesn't make too much sense for Wasm; maybe just omit this method there?
     fn get_default_window_size(&self) -> Vec2;
 
-    /// Request for a file to be read. Makes some effort to defer actually reading of the file until
-    /// later. Currently the behavior is as follows:
-    /// * Wasm32 target: Always loads the entire file over HTTP(S).
-    /// * Native targets: If `path` is a relative file name, we don't load the file but instead open
-    //    a read handle. If `path` is an HTTP URL we load the whole file. HTTPS is not supported yet.
-    fn file_read(&mut self, path: &str) -> FileRead;
-
-    /// Get an actual file reader based on a [`UniversalFileHandle`].
-    fn get_file_reader<'a>(&mut self, universal_file_handle: &'a UniversalFileHandle) -> Result<Box<dyn ReadSeek + 'a>, String>;
-
-    /// Get a file reader, but consumes the UniversalFileHandle if it contains
-    /// [`UniversalFileHandle::Data`]. Useful if you want to store the resulting reader somewhere.
-    fn get_file_reader_consume(&mut self, universal_file_handle: &mut UniversalFileHandle) -> Result<Box<dyn ReadSeek>, String>;
-
     /// Write `data` to `path`.
     fn file_write(&mut self, path: &str, data: &[u8]);
 
@@ -859,7 +895,7 @@ pub trait CxDesktopVsWasmCommon {
     fn websocket_send(&mut self, url: &str, data: &[u8]);
 
     /// Make an HTTP request. When done, you get a [`SignalEvent`] corresponding to the provided
-    /// [`Signal`] with [`Cx::status_http_send_ok`] or [`Cx::status_http_send_fail`] as the status.
+    /// [`Signal`] with [`Cx::STATUS_HTTP_SEND_OK`] or [`Cx::STATUS_HTTP_SEND_FAIL`] as the status.
     fn http_send(
         &mut self,
         verb: &str,
@@ -871,6 +907,18 @@ pub trait CxDesktopVsWasmCommon {
         body: &[u8],
         signal: Signal,
     );
+
+    /// Call JS function from Rust. Must be called on main thread and call a function already
+    /// registered using `register_call_js_callbacks`. `params` is an arbitrary string.
+    /// `buffers` is an array of reference-counted byte buffers.
+    #[cfg(any(target_arch = "wasm32", feature = "cef"))]
+    fn call_js(&mut self, name: &str, params: Vec<ZapParam>);
+
+    /// Mechanism to communicate back returns values from call_rust functions.
+    fn return_to_js(&mut self, callback_id: u32, params: Vec<ZapParam>);
+
+    #[cfg(any(target_arch = "wasm32", feature = "cef"))]
+    fn on_call_rust_in_same_thread_sync(&mut self, func: CallRustInSameThreadSyncFn);
 }
 
 /// A bunch of traits that are common between the different target platforms. This trait makes sure
@@ -883,11 +931,15 @@ pub trait CxPlatformCommon {
     fn hide_text_ime(&mut self);
     /// Start a new [`Timer`] with the given `interval`, and which may `repeat` if required.
     fn start_timer(&mut self, interval: f64, repeats: bool) -> Timer;
-    /// Stop a [`Timer`] given by [`CxPlatformCommon::start_timer`].
+    /// Stop zap`Timer`] given by [`CxPlatformCommon::start_timer`].
     fn stop_timer(&mut self, timer: &mut Timer);
     /// Post a [`Signal`] from any thread. If you don't need to use this from a thread, you may
     /// instead use [`Cx::send_signal`], which might be faster.
     fn post_signal(signal: Signal, status: StatusId);
     /// Set a [`Menu`].
     fn update_menu(&mut self, menu: &Menu);
+    /// Copy the given text to the clipboard, if possible.
+    fn copy_text_to_clipboard(&mut self, text: &str);
+    /// Send zaplib Event for processing from any thread
+    fn send_event_from_any_thread(event: Event);
 }
