@@ -1,9 +1,14 @@
 //! Bindings for OSX Core Foundation and Metal.
 
+// Clippy TODO
+#![allow(clippy::enum_variant_names)]
 #![allow(dead_code)]
 #![allow(non_camel_case_types)]
 #![allow(non_upper_case_globals)]
 
+use std::ffi::c_void;
+use std::ptr::NonNull;
+
 pub(crate) use zaplib_objc_sys::declare::ClassDecl;
 pub(crate) use zaplib_objc_sys::runtime::{Class, Object, Protocol, Sel, BOOL, NO, YES};
 pub(crate) use zaplib_objc_sys::{class, msg_send, sel, sel_impl};
@@ -13,23 +18,57 @@ pub(crate) use zaplib_objc_sys::{Encode, Encoding};
 pub(crate) type id = *mut zaplib_objc_sys::runtime::Object;
 pub(crate) const nil: id = 0 as id;
 
+pub struct RcObjcId(NonNull<Object>);
+
+impl RcObjcId {
+    pub fn from_owned(id: NonNull<Object>) -> Self {
+        Self(id)
+    }
+
+    pub fn from_unowned(id: NonNull<Object>) -> Self {
+        unsafe {
+            let _: () = msg_send![id.as_ptr(), retain];
+        }
+        Self::from_owned(id)
+    }
+
+    pub fn as_id(&self) -> id {
+        self.0.as_ptr()
+    }
+
+    pub fn forget(self) -> NonNull<Object> {
+        unsafe {
+            let _: () = msg_send![self.0.as_ptr(), retain];
+        }
+        self.0
+    }
+}
+
+#[link(name = "system")]
+extern "C" {
+    pub static _NSConcreteStackBlock: [*const c_void; 32];
+    pub static _NSConcreteBogusBlock: [*const c_void; 32];
+}
+
 #[link(name = "Foundation", kind = "framework")]
 extern "C" {
     pub(crate) static NSRunLoopCommonModes: id;
     pub(crate) static NSDefaultRunLoopMode: id;
+    pub(crate) fn NSStringFromClass(cls: id) -> id;
 }
 
 #[link(name = "AppKit", kind = "framework")]
 extern "C" {
     pub(crate) static NSStringPboardType: id;
     pub(crate) static NSFilenamesPboardType: id;
+    pub(crate) static NSURLPboardType: id;
 }
 
 #[link(name = "CoreGraphics", kind = "framework")]
 extern "C" {
     pub(crate) fn CGMainDisplayID() -> u32;
     pub(crate) fn CGDisplayPixelsHigh(display: u32) -> u64;
-    pub(crate) fn CGColorCreateGenericRGB(red: f64, green: f64, blue: f64, alpha: f64) -> id;
+    pub(crate) fn CGColorCreateSRGB(red: f64, green: f64, blue: f64, alpha: f64) -> id;
 }
 
 #[link(name = "Metal", kind = "framework")]
@@ -142,7 +181,7 @@ pub(crate) fn load_webkit_cursor(cursor_name_str: &str) -> id {
 }
 
 #[repr(C)]
-#[derive(Copy, Clone)]
+#[derive(Copy, Clone, Debug)]
 pub(crate) struct NSPoint {
     pub(crate) x: f64,
     pub(crate) y: f64,
@@ -424,6 +463,7 @@ pub(crate) struct MTLClearColor {
 #[allow(non_camel_case_types)]
 #[derive(Debug, Copy, Clone, Hash, PartialEq, Eq)]
 pub(crate) enum MTLPixelFormat {
+    RGBA8Unorm = 70,
     BGRA8Unorm = 80,
     Depth32Float = 252,
     Stencil8 = 253,
