@@ -2,8 +2,8 @@
 
 #![allow(dead_code)]
 
-use crate::cx::*;
 use crate::cx_win32::*;
+use crate::*;
 use std::ffi;
 use std::mem;
 use std::ptr;
@@ -50,15 +50,17 @@ impl Cx {
                     zbias_step,
                 );
             } else {
+                let gpu_geometry_id = GpuGeometry::get_id(self, view_id, draw_call_id);
+
                 let cxview = &mut self.views[view_id];
 
                 let draw_call = &mut cxview.draw_calls[draw_call_id];
-                let sh = &self.shaders[draw_call.shader.shader_id];
+                let sh = &self.shaders[draw_call.shader_id];
                 let shp = sh.platform.as_ref().unwrap();
 
                 if draw_call.instance_dirty {
                     draw_call.instance_dirty = false;
-                    if draw_call.instances.len() == 0 {
+                    if draw_call.instances.is_empty() {
                         continue;
                     }
                     // update the instance buffer data
@@ -75,7 +77,7 @@ impl Cx {
 
                 if draw_call.uniforms_dirty {
                     draw_call.uniforms_dirty = false;
-                    if draw_call.user_uniforms.len() != 0 {
+                    if !draw_call.user_uniforms.is_empty() {
                         draw_call.platform.user_uniforms.update_with_f32_constant_data(d3d11_cx, &mut draw_call.user_uniforms);
                     }
                 }
@@ -86,19 +88,11 @@ impl Cx {
                     continue;
                 }
 
-                let geometry_id = if let Some(geometry) = draw_call.geometry {
-                    geometry.geometry_id
-                } else if let Some(geometry) = sh.default_geometry {
-                    geometry.geometry_id
-                } else {
-                    continue;
-                };
-
-                let geometry = &mut self.geometries[geometry_id];
+                let geometry = &mut self.gpu_geometries[gpu_geometry_id];
 
                 if geometry.dirty {
-                    geometry.platform.geom_ibuf.update_with_u32_index_data(d3d11_cx, &geometry.indices);
-                    geometry.platform.geom_vbuf.update_with_f32_vertex_data(d3d11_cx, &geometry.vertices);
+                    geometry.platform.geom_vbuf.update_with_f32_vertex_data(d3d11_cx, geometry.geometry.vertices_f32_slice());
+                    geometry.platform.geom_ibuf.update_with_u32_index_data(d3d11_cx, geometry.geometry.indices_u32_slice());
                     geometry.dirty = false;
                 }
 
@@ -128,10 +122,10 @@ impl Cx {
                     let cxtexture = &mut self.textures[*texture_id as usize];
                     match cxtexture.desc.format {
                         // we only allocate Image and Mapped textures.
-                        TextureFormat::Default | TextureFormat::ImageBGRA => {
+                        TextureFormat::ImageRGBA => {
                             if cxtexture.update_image {
                                 cxtexture.update_image = false;
-                                d3d11_cx.update_platform_texture_image_bgra(
+                                d3d11_cx.update_platform_texture_image_rgba(
                                     &mut cxtexture.platform,
                                     cxtexture.desc.width.unwrap(),
                                     cxtexture.desc.height.unwrap(),
@@ -140,18 +134,11 @@ impl Cx {
                             }
                             d3d11_cx.set_shader_resource(i, &cxtexture.platform.shader_resource);
                         }
-
-                        TextureFormat::RenderBGRA | TextureFormat::RenderBGRAf16 | TextureFormat::RenderBGRAf32 => {
-                            d3d11_cx.set_shader_resource(i, &cxtexture.platform.shader_resource);
-                        }
                         _ => (),
                     }
                 }
-                //if self.passes[pass_id].debug{
-                //    println!("DRAWING PASS {} {}", geometry.indices.len(), instances);
-                //}
 
-                d3d11_cx.draw_indexed_instanced(geometry.indices.len(), instances);
+                d3d11_cx.draw_indexed_instanced(geometry.geometry.indices_u32_slice().len(), instances);
             }
         }
         self.debug_draw_tree(view_id);
@@ -197,12 +184,12 @@ impl Cx {
             match color_texture.clear_color {
                 ClearColor::InitWith(color) => {
                     if is_initial {
-                        d3d11_cx.clear_render_target_view(&render_target, color);
+                        d3d11_cx.clear_render_target_view(render_target, color);
                         //self.clear_color);
                     }
                 }
                 ClearColor::ClearWith(color) => {
-                    d3d11_cx.clear_render_target_view(&render_target, color); //self.clear_color);
+                    d3d11_cx.clear_render_target_view(render_target, color); //self.clear_color);
                 }
             }
         }
@@ -248,7 +235,7 @@ impl Cx {
         d3d11_cx.set_raster_state(self.passes[pass_id].platform.raster_state.as_ref().unwrap());
         d3d11_cx.set_blend_state(self.passes[pass_id].platform.blend_state.as_ref().unwrap());
         let cxpass = &mut self.passes[pass_id];
-        cxpass.platform.pass_uniforms.update_with_f32_constant_data(&d3d11_cx, cxpass.pass_uniforms.as_slice());
+        cxpass.platform.pass_uniforms.update_with_f32_constant_data(d3d11_cx, cxpass.pass_uniforms.as_slice());
     }
 
     pub(crate) fn draw_pass_to_window(
@@ -270,7 +257,7 @@ impl Cx {
             view_id,
             Vec2::default(),
             (Vec2 { x: -50000., y: -50000. }, Vec2 { x: 50000., y: 50000. }),
-            &d3d11_cx,
+            d3d11_cx,
             &mut zbias,
             zbias_step,
         );
@@ -289,20 +276,17 @@ impl Cx {
             view_id,
             Vec2::default(),
             (Vec2 { x: -50000., y: -50000. }, Vec2 { x: 50000., y: 50000. }),
-            &d3d11_cx,
+            d3d11_cx,
             &mut zbias,
             zbias_step,
         );
     }
 
-    pub(crate) const fn use_metal_uniform_packing() -> bool {
-        false
-    }
-
     pub(crate) fn hlsl_compile_shaders(&mut self, d3d11_cx: &D3d11Cx) {
-        for shader in &mut self.shaders {
+        for shader_id in self.shader_recompile_ids.drain(..) {
+            let shader = unsafe { self.shaders.get_unchecked_mut(shader_id) };
             let shader_ast = shader.shader_ast.as_ref().unwrap();
-            let hlsl = generate_hlsl::generate_shader(&shader_ast);
+            let hlsl = generate_hlsl::generate_shader(shader_ast);
             let debug = shader_ast.debug;
             if debug {
                 println!("--------------- Shader {} --------------- \n{}\n", &shader.name, hlsl);
@@ -312,14 +296,14 @@ impl Cx {
 
             fn split_source(src: &str) -> String {
                 let mut r = String::new();
-                let split = src.split("\n");
+                let split = src.split('\n');
                 for (line, chunk) in split.enumerate() {
                     r.push_str(&(line + 1).to_string());
-                    r.push_str(":");
+                    r.push(':');
                     r.push_str(chunk);
-                    r.push_str("\n");
+                    r.push('\n');
                 }
-                return r;
+                r
             }
 
             if let Err(msg) = vs_blob {
@@ -659,7 +643,7 @@ impl D3d11Cx {
             if let Some(uni) = user_uni.buffer.as_ref() {
                 uni.as_raw() as *const std::ffi::c_void
             } else {
-                0 as *const std::ffi::c_void
+                std::ptr::null::<std::ffi::c_void>()
             },
         ];
         unsafe { self.context.VSSetConstantBuffers(0, 6, buffers.as_ptr() as *const *mut _) };
@@ -999,15 +983,9 @@ impl D3d11Cx {
 
         let format;
         match cxtexture.desc.format {
-            TextureFormat::Default | TextureFormat::RenderBGRA => {
+            TextureFormat::ImageRGBA => {
                 format = dxgiformat::DXGI_FORMAT_R8G8B8A8_UNORM;
             }
-            TextureFormat::RenderBGRAf16 => {
-                format = dxgiformat::DXGI_FORMAT_R32G32B32A32_FLOAT;
-            }
-            TextureFormat::RenderBGRAf32 => {
-                format = dxgiformat::DXGI_FORMAT_R32G32B32A32_FLOAT;
-            }
             _ => {
                 panic!("Wrong format for update_render_target");
             }
@@ -1046,7 +1024,7 @@ impl D3d11Cx {
         } else {
             panic!("update_render_target failed");
         }
-        return true;
+        true
     }
 
     pub(crate) fn update_depth_stencil(&self, cxtexture: &mut CxTexture, dpi_factor: f32, size: Vec2) -> bool {
@@ -1059,7 +1037,7 @@ impl D3d11Cx {
 
         let format;
         match cxtexture.desc.format {
-            TextureFormat::Default | TextureFormat::Depth32Stencil8 => {
+            TextureFormat::Depth32Stencil8 => {
                 format = dxgiformat::DXGI_FORMAT_D32_FLOAT_S8X24_UINT;
             }
             _ => {
@@ -1098,10 +1076,10 @@ impl D3d11Cx {
         } else {
             panic!("update_render_target failed");
         }
-        return true;
+        true
     }
 
-    pub(crate) fn update_platform_texture_image_bgra(
+    pub(crate) fn update_platform_texture_image_rgba(
         &self,
         res: &mut CxPlatformTexture,
         width: usize,
@@ -1109,7 +1087,7 @@ impl D3d11Cx {
         image_u32: &Vec<u32>,
     ) {
         if image_u32.len() != width * height {
-            println!("update_platform_texture_image_bgra with wrong buffer_u32 size!");
+            println!("update_platform_texture_image_rgba with wrong buffer_u32 size!");
             return;
         }
 
@@ -1141,7 +1119,7 @@ impl D3d11Cx {
             res.texture = Some(unsafe { ComPtr::from_raw(texture as *mut _) });
             res.shader_resource = Some(unsafe { ComPtr::from_raw(shader_resource as *mut _) });
         } else {
-            panic!("update_platform_texture_image_bgra failed");
+            panic!("update_platform_texture_image_rgba failed");
         }
     }
 }
@@ -1257,7 +1235,7 @@ pub(crate) struct CxPlatformPass {
 }
 
 #[derive(Default, Clone)]
-pub(crate) struct CxPlatformGeometry {
+pub(crate) struct CxPlatformGpuGeometry {
     pub(crate) geom_vbuf: D3d11Buffer,
     pub(crate) geom_ibuf: D3d11Buffer,
 }
