@@ -1,7 +1,7 @@
 //! Linux OpenGL bindings.
 
-use crate::cx::*;
 use crate::cx_xlib::*;
+use crate::*;
 use std::ffi::{CStr, CString};
 use std::mem;
 use std::os::raw::{c_ulong, c_void};
@@ -41,10 +41,12 @@ impl Cx {
                     zbias_step,
                 );
             } else {
+                let gpu_geometry_id = GpuGeometry::get_id(self, view_id, draw_call_id);
+
                 let cxview = &mut self.views[view_id];
                 //view.platform.uni_vw.update_with_f32_data(device, &view.uniforms);
                 let draw_call = &mut cxview.draw_calls[draw_call_id];
-                let sh = &self.shaders[draw_call.shader.shader_id];
+                let sh = &self.shaders[draw_call.shader_id];
                 let shp = sh.platform.as_ref().unwrap();
 
                 if draw_call.instance_dirty {
@@ -52,16 +54,8 @@ impl Cx {
                     draw_call.platform.inst_vb.update_with_f32_data(opengl_cx, &draw_call.instances);
                 }
 
-                let geometry_id = if let Some(geometry) = draw_call.geometry {
-                    geometry.geometry_id
-                } else if let Some(geometry) = sh.default_geometry {
-                    geometry.geometry_id
-                } else {
-                    continue;
-                };
-
-                let geometry = &mut self.geometries[geometry_id];
-                let indices = geometry.indices.len();
+                let geometry = &mut self.gpu_geometries[gpu_geometry_id];
+                let indices = geometry.geometry.indices_u32_slice().len();
 
                 draw_call.set_zbias(*zbias);
                 draw_call.set_local_scroll(scroll, local_scroll);
@@ -73,10 +67,10 @@ impl Cx {
                 }
 
                 // update geometry?
-                let geometry = &mut self.geometries[geometry_id];
+                let geometry = &mut self.gpu_geometries[gpu_geometry_id];
                 if geometry.dirty || geometry.platform.vb.gl_buffer.is_none() || geometry.platform.ib.gl_buffer.is_none() {
-                    geometry.platform.vb.update_with_f32_data(opengl_cx, &geometry.vertices);
-                    geometry.platform.ib.update_with_u32_data(opengl_cx, &geometry.indices);
+                    geometry.platform.vb.update_with_f32_data(opengl_cx, geometry.geometry.vertices_f32_slice());
+                    geometry.platform.ib.update_with_u32_data(opengl_cx, geometry.geometry.indices_u32_slice());
                     geometry.dirty = false;
                 }
 
@@ -99,9 +93,9 @@ impl Cx {
                 if vao.inst_vb != draw_call.platform.inst_vb.gl_buffer
                     || vao.geom_vb != geometry.platform.vb.gl_buffer
                     || vao.geom_ib != geometry.platform.ib.gl_buffer
-                    || vao.shader_id != Some(draw_call.shader.shader_id)
+                    || vao.shader_id != Some(draw_call.shader_id)
                 {
-                    vao.shader_id = Some(draw_call.shader.shader_id);
+                    vao.shader_id = Some(draw_call.shader_id);
                     vao.inst_vb = draw_call.platform.inst_vb.gl_buffer;
                     vao.geom_vb = geometry.platform.vb.gl_buffer;
                     vao.geom_ib = geometry.platform.ib.gl_buffer;
@@ -135,7 +129,7 @@ impl Cx {
                                 attr.offset as *const () as *const _,
                             );
                             gl::EnableVertexAttribArray(attr.loc);
-                            gl::VertexAttribDivisor(attr.loc, 1 as gl::types::GLuint);
+                            gl::VertexAttribDivisor(attr.loc, 1_u32);
                         }
 
                         // bind the indexbuffer
@@ -209,12 +203,10 @@ impl Cx {
             opengl_window.opening_repaint_count += 1;
             init_repaint = true;
         }
-        let window;
-        let view_rect;
 
         opengl_window.xlib_window.hide_child_windows();
 
-        window = opengl_window.xlib_window.window.unwrap();
+        let window = opengl_window.xlib_window.window.unwrap();
 
         let pass_size = self.passes[pass_id].pass_size;
         self.passes[pass_id].set_matrix(Vec2::default(), pass_size);
@@ -226,12 +218,12 @@ impl Cx {
             glx_sys::glXMakeCurrent(opengl_cx.display, window, opengl_cx.context);
             gl::Viewport(0, 0, pix_width as i32, pix_height as i32);
         }
-        view_rect = Rect::default();
+        let view_rect = Rect::default();
 
         //self.passes[pass_id].uniform_camera_view(&Mat4::identity());
         self.passes[pass_id].set_dpi_factor(dpi_factor);
         // set up the
-        let clear_color = if self.passes[pass_id].color_textures.len() == 0 {
+        let clear_color = if self.passes[pass_id].color_textures.is_empty() {
             Vec4::default()
         } else {
             match self.passes[pass_id].color_textures[0].clear_color {
@@ -261,7 +253,7 @@ impl Cx {
             Vec2::default(),
             (Vec2 { x: -50000., y: -50000. }, Vec2 { x: 50000., y: 50000. }),
             &view_rect,
-            &opengl_cx,
+            opengl_cx,
             &mut zbias,
             zbias_step,
         );
@@ -269,7 +261,7 @@ impl Cx {
         unsafe {
             glx_sys::glXSwapBuffers(opengl_cx.display, window);
         }
-        return init_repaint;
+        init_repaint
     }
 
     pub(crate) fn draw_pass_to_texture(&mut self, pass_id: usize, inherit_dpi_factor: f32, opengl_cx: &OpenglCx) {
@@ -423,7 +415,7 @@ impl Cx {
             Vec2::default(),
             (Vec2 { x: -50000., y: -50000. }, Vec2 { x: 50000., y: 50000. }),
             &Rect::default(),
-            &opengl_cx,
+            opengl_cx,
             &mut zbias,
             zbias_step,
         );
@@ -479,13 +471,13 @@ impl Cx {
             log.set_len(length as usize);
             let mut r = "".to_string();
             r.push_str(CStr::from_ptr(log.as_ptr()).to_str().unwrap());
-            r.push_str("\n");
-            let split = source.split("\n");
+            r.push('\n');
+            let split = source.split('\n');
             for (line, chunk) in split.enumerate() {
                 r.push_str(&(line + 1).to_string());
-                r.push_str(":");
+                r.push(':');
                 r.push_str(chunk);
-                r.push_str("\n");
+                r.push('\n');
             }
             r
         }
@@ -526,7 +518,7 @@ impl Cx {
         for i in 0..num_attr {
             let mut name0 = prefix.to_string();
             name0.push_str(&i.to_string());
-            name0.push_str("\0");
+            name0.push('\0');
 
             let mut size = (slots - i * 4) as i32;
             if size > 4 {
@@ -534,10 +526,7 @@ impl Cx {
             }
             unsafe {
                 attribs.push(OpenglAttribute {
-                    loc: {
-                        let loc = gl::GetAttribLocation(program, name0.as_ptr() as *const _) as u32;
-                        loc
-                    },
+                    loc: gl::GetAttribLocation(program, name0.as_ptr() as *const _) as u32,
                     offset: (i * 4 * mem::size_of::<f32>()) as usize,
                     size,
                     stride,
@@ -547,7 +536,7 @@ impl Cx {
         attribs
     }
 
-    pub(crate) fn opengl_get_uniforms(program: u32, unis: &Vec<PropDef>) -> Vec<OpenglUniform> {
+    pub(crate) fn opengl_get_uniforms(program: u32, unis: &[PropDef]) -> Vec<OpenglUniform> {
         let mut gl_uni = Vec::new();
         for uni in unis {
             gl_uni.push(Self::opengl_get_uniform(program, &uni.name, uni.ty.size()));
@@ -558,26 +547,24 @@ impl Cx {
     pub(crate) fn opengl_get_uniform(program: u32, name: &str, size: usize) -> OpenglUniform {
         let mut name0 = String::new();
         name0.push_str(name);
-        name0.push_str("\0");
-        unsafe {
-            OpenglUniform { loc: gl::GetUniformLocation(program, name0.as_ptr() as *const _), name: name.to_string(), size }
-        }
-    }
-
-    pub(crate) const fn use_metal_uniform_packing() -> bool {
-        false
+        name0.push('\0');
+        unsafe { OpenglUniform { loc: gl::GetUniformLocation(program, name0.as_ptr() as *const _), size } }
     }
 
     pub(crate) fn opengl_compile_shaders(&mut self, opengl_cx: &OpenglCx) {
+        if self.shader_recompile_ids.is_empty() {
+            return;
+        }
+
         unsafe {
             glx_sys::glXMakeCurrent(opengl_cx.display, opengl_cx.hidden_window, opengl_cx.context);
         }
-
-        for shader in &mut self.shaders {
+        for shader_id in self.shader_recompile_ids.drain(..) {
+            let shader = unsafe { self.shaders.get_unchecked_mut(shader_id) };
             let shader_ast = shader.shader_ast.as_ref().unwrap();
 
-            let vertex = generate_glsl::generate_vertex_shader(&shader_ast);
-            let fragment = generate_glsl::generate_fragment_shader(&shader_ast);
+            let vertex = generate_glsl::generate_vertex_shader(shader_ast);
+            let fragment = generate_glsl::generate_fragment_shader(shader_ast);
 
             let vertex = format!(
                 "
@@ -640,8 +627,6 @@ impl Cx {
                     program,
                     geometries,
                     instances,
-                    vertex,
-                    fragment,
                     pass_uniforms: Self::opengl_get_uniforms(program, &shader.mapping.pass_uniforms),
                     view_uniforms: Self::opengl_get_uniforms(program, &shader.mapping.view_uniforms),
                     draw_uniforms: Self::opengl_get_uniforms(program, &shader.mapping.draw_uniforms),
@@ -778,7 +763,7 @@ impl OpenglCx {
         }
     }
 
-    pub(crate) fn set_uniform_buffer(&self, locs: &Vec<OpenglUniform>, uni: &[f32]) {
+    pub(crate) fn set_uniform_buffer(&self, locs: &[OpenglUniform], uni: &[f32]) {
         let mut o = 0;
         for loc in locs {
             if o + loc.size > uni.len() {
@@ -800,13 +785,13 @@ impl OpenglCx {
                             gl::Uniform4f(loc.loc as i32, uni[o], uni[o + 1], uni[o + 2], uni[o + 3]);
                         }
                         16 => {
-                            gl::UniformMatrix4fv(loc.loc as i32, 1, 0, uni.as_ptr().offset((o) as isize));
+                            gl::UniformMatrix4fv(loc.loc as i32, 1, 0, uni.as_ptr().add(o));
                         }
                         _ => (),
                     }
                 }
             };
-            o = o + loc.size;
+            o += loc.size;
         }
     }
 
@@ -820,6 +805,8 @@ impl OpenglCx {
         let height = cxtexture.desc.height.unwrap();
 
         // allocate new texture if descriptor change
+        // TODO(Paras): Change this so that we still update image_u32 data
+        // whenever this function is called, not just when the descriptor changes.
         if cxtexture.platform.alloc_desc != cxtexture.desc {
             cxtexture.platform.alloc_desc = cxtexture.desc.clone();
             cxtexture.platform.width = width as u64;
@@ -890,7 +877,7 @@ impl OpenglCx {
 
             if !is_depth {
                 match cxtexture.desc.format {
-                    TextureFormat::Default | TextureFormat::RenderBGRA => {
+                    TextureFormat::ImageRGBA => {
                         let mut gl_texture = std::mem::MaybeUninit::uninit();
                         gl::GenTextures(1, gl_texture.as_mut_ptr());
                         let gl_texture = gl_texture.assume_init();
@@ -919,7 +906,7 @@ impl OpenglCx {
                 }
             } else {
                 match cxtexture.desc.format {
-                    TextureFormat::Default | TextureFormat::Depth32Stencil8 => {
+                    TextureFormat::Depth32Stencil8 => {
                         let mut gl_renderbuf = std::mem::MaybeUninit::uninit();
                         gl::GenRenderbuffers(1, gl_renderbuf.as_mut_ptr());
                         let gl_renderbuffer = gl_renderbuf.assume_init();
@@ -929,21 +916,19 @@ impl OpenglCx {
                         cxtexture.platform.gl_renderbuffer = Some(gl_renderbuffer);
                     }
                     _ => {
-                        println!("update_platform_render_targete unsupported texture format");
+                        println!("update_platform_render_target unsupported texture format");
                         return false;
                     }
                 }
             }
         }
-        return true;
+        true
     }
 }
 
 #[derive(Clone)]
 pub(crate) struct CxPlatformShader {
     pub(crate) program: u32,
-    pub(crate) vertex: String,
-    pub(crate) fragment: String,
     pub(crate) geometries: Vec<OpenglAttribute>,
     pub(crate) instances: Vec<OpenglAttribute>,
     pub(crate) pass_uniforms: Vec<OpenglUniform>,
@@ -1012,12 +997,11 @@ pub(crate) struct OpenglAttribute {
 #[derive(Debug, Default, Clone)]
 pub(crate) struct OpenglUniform {
     pub(crate) loc: i32,
-    pub(crate) name: String,
     pub(crate) size: usize,
 }
 
 #[derive(Clone, Default)]
-pub(crate) struct CxPlatformGeometry {
+pub(crate) struct CxPlatformGpuGeometry {
     pub(crate) vb: OpenglBuffer,
     pub(crate) ib: OpenglBuffer,
 }
@@ -1078,7 +1062,7 @@ impl OpenglBuffer {
         }
     }
 
-    pub(crate) fn update_with_f32_data(&mut self, _opengl_cx: &OpenglCx, data: &Vec<f32>) {
+    pub(crate) fn update_with_f32_data(&mut self, _opengl_cx: &OpenglCx, data: &[f32]) {
         if self.gl_buffer.is_none() {
             self.alloc_gl_buffer();
         }
@@ -1093,7 +1077,7 @@ impl OpenglBuffer {
         }
     }
 
-    pub(crate) fn update_with_u32_data(&mut self, _opengl_cx: &OpenglCx, data: &Vec<u32>) {
+    pub(crate) fn update_with_u32_data(&mut self, _opengl_cx: &OpenglCx, data: &[u32]) {
         if self.gl_buffer.is_none() {
             self.alloc_gl_buffer();
         }
