@@ -1,7 +1,8 @@
 //! Win32 API bindings.
 
-use crate::cx::*;
+use crate::*;
 use std::collections::{BTreeSet, HashMap};
+use std::convert::TryInto;
 use std::ffi::OsStr;
 use std::mem;
 use std::os::raw::c_void;
@@ -37,8 +38,6 @@ pub(crate) struct Win32App {
     pub(crate) loop_block: bool,
     pub(crate) dpi_functions: DpiFunctions,
     pub(crate) current_cursor: MouseCursor,
-    pub(crate) status_to_usize: HashMap<StatusId, usize>,
-    pub(crate) usize_to_status: HashMap<usize, StatusId>,
 }
 
 #[derive(Clone)]
@@ -49,12 +48,9 @@ pub(crate) struct Win32Window {
 
     pub(crate) time_start: u64,
 
-    pub(crate) last_key_mod: KeyModifiers,
     pub(crate) ime_spot: Vec2,
-    pub(crate) current_cursor: MouseCursor,
     pub(crate) last_mouse_pos: Vec2,
-    pub(crate) fingers_down: Vec<bool>,
-    pub(crate) ignore_wmsize: usize,
+    pub(crate) pointers_down: Vec<bool>,
     pub(crate) hwnd: Option<HWND>,
     pub(crate) track_mouse_event: bool,
 }
@@ -62,7 +58,7 @@ pub(crate) struct Win32Window {
 #[derive(Clone)]
 pub(crate) enum Win32Timer {
     Free,
-    Timer { win32_id: UINT_PTR, timer_id: u64, interval: f64, repeats: bool },
+    Timer { win32_id: UINT_PTR, timer_id: u64, repeats: bool },
     Resize { win32_id: UINT_PTR },
 }
 
@@ -103,8 +99,6 @@ impl Win32App {
             free_timers: Vec::new(),
             dpi_functions: DpiFunctions::new(),
             current_cursor: MouseCursor::Default,
-            status_to_usize: HashMap::new(),
-            usize_to_status: HashMap::new(),
         };
 
         win32_app.dpi_functions.become_dpi_aware();
@@ -118,6 +112,8 @@ impl Win32App {
         }
     }
 
+    /// TODO(JP): We learned a lot in improving the OSX event loop and will likely need to apply
+    /// some of the things we did there to here.
     pub(crate) fn event_loop<F>(&mut self, mut event_handler: F)
     where
         F: FnMut(&mut Win32App, &mut Vec<Event>) -> bool,
@@ -140,14 +136,14 @@ impl Win32App {
                     } else {
                         winuser::TranslateMessage(&msg);
                         winuser::DispatchMessageW(&msg);
-                        self.do_callback(&mut vec![Event::Paint]);
+                        self.do_callback(&mut vec![Event::SystemEvent(SystemEvent::Paint)]);
                     }
                 } else {
                     let mut msg = std::mem::MaybeUninit::uninit();
                     let ret = winuser::PeekMessageW(msg.as_mut_ptr(), ptr::null_mut(), 0, 0, 1);
                     let msg = msg.assume_init();
                     if ret == 0 {
-                        self.do_callback(&mut vec![Event::Paint])
+                        self.do_callback(&mut vec![Event::SystemEvent(SystemEvent::Paint)])
                     } else {
                         winuser::TranslateMessage(&msg);
                         winuser::DispatchMessageW(&msg);
@@ -205,7 +201,7 @@ impl Win32App {
                     win32_app.do_callback(&mut vec![Event::Timer(TimerEvent { timer_id })]);
                 }
                 Win32Timer::Resize { .. } => {
-                    win32_app.do_callback(&mut vec![Event::Paint]);
+                    win32_app.do_callback(&mut vec![Event::SystemEvent(SystemEvent::Paint)]);
                 }
                 _ => (),
             }
@@ -213,7 +209,7 @@ impl Win32App {
     }
 
     pub(crate) fn get_free_timer_slot(&mut self) -> usize {
-        if self.free_timers.len() > 0 {
+        if !self.free_timers.is_empty() {
             self.free_timers.pop().unwrap()
         } else {
             let slot = self.timers.len();
@@ -225,7 +221,7 @@ impl Win32App {
     pub(crate) fn start_timer(&mut self, timer_id: u64, interval: f64, repeats: bool) {
         let slot = self.get_free_timer_slot();
         let win32_id = unsafe { winuser::SetTimer(NULL as HWND, 0, (interval * 1000.0) as u32, Some(Self::timer_proc)) };
-        self.timers[slot] = Win32Timer::Timer { timer_id, win32_id, interval, repeats };
+        self.timers[slot] = Win32Timer::Timer { timer_id, win32_id, repeats };
     }
 
     pub(crate) fn stop_timer(&mut self, which_timer_id: u64) {
@@ -244,7 +240,7 @@ impl Win32App {
 
     pub(crate) fn start_resize(&mut self) {
         let slot = self.get_free_timer_slot();
-        let win32_id = unsafe { winuser::SetTimer(NULL as HWND, 0, 8 as u32, Some(Self::timer_proc)) };
+        let win32_id = unsafe { winuser::SetTimer(NULL as HWND, 0, 8_u32, Some(Self::timer_proc)) };
         self.timers[slot] = Win32Timer::Resize { win32_id };
     }
 
@@ -263,17 +259,10 @@ impl Win32App {
     pub(crate) fn post_signal(signal: Signal, status: StatusId) {
         unsafe {
             let win32_app = &mut (*GLOBAL_WIN32_APP);
-            if let Ok(mut sigs) = win32_app.race_signals.lock() {
-                let status_id = if let Some(id) = win32_app.status_to_usize.get(&status) {
-                    *id
-                } else {
-                    let id = win32_app.status_to_usize.len();
-                    win32_app.status_to_usize.insert(status, id);
-                    win32_app.usize_to_status.insert(id, status);
-                    id
-                };
+            let status_id = isize::from_ne_bytes(status.0.to_ne_bytes());
 
-                if win32_app.all_windows.len() > 0 {
+            if let Ok(mut sigs) = win32_app.race_signals.lock() {
+                if !win32_app.all_windows.is_empty() {
                     winuser::PostMessageW(
                         win32_app.all_windows[0],
                         winuser::WM_USER,
@@ -290,7 +279,7 @@ impl Win32App {
 
     pub(crate) fn terminate_event_loop(&mut self) {
         unsafe {
-            if self.all_windows.len() > 0 {
+            if !self.all_windows.is_empty() {
                 winuser::PostMessageW(self.all_windows[0], winuser::WM_QUIT, 0, 0);
             }
         }
@@ -345,24 +334,43 @@ impl Win32App {
             //TODO
         }
     }
+
+    pub(crate) fn copy_text_to_clipboard(text: &str) {
+        // plug it into the windows clipboard
+        // make utf16 dta
+        unsafe {
+            if winuser::OpenClipboard(ptr::null_mut()) != 0 {
+                winuser::EmptyClipboard();
+
+                let data: Vec<u16> = OsStr::new(text).encode_wide().chain(Some(0).into_iter()).collect();
+
+                let h_clipboard_data = winbase::GlobalAlloc(winbase::GMEM_DDESHARE, 2 * data.len());
+
+                let h_clipboard_ptr = winbase::GlobalLock(h_clipboard_data) as *mut u16;
+
+                std::ptr::copy_nonoverlapping(data.as_ptr(), h_clipboard_ptr, data.len());
+
+                winbase::GlobalUnlock(h_clipboard_data);
+                winuser::SetClipboardData(winuser::CF_UNICODETEXT, h_clipboard_data);
+                winuser::CloseClipboard();
+            }
+        }
+    }
 }
 
 impl Win32Window {
     pub(crate) fn new(win32_app: &mut Win32App, window_id: usize) -> Win32Window {
-        let mut fingers_down = Vec::new();
-        fingers_down.resize(NUM_FINGERS, false);
+        let mut pointers_down = Vec::new();
+        pointers_down.resize(NUM_POINTERS, false);
 
         Win32Window {
             window_id,
             win32_app,
             last_window_geom: WindowGeom::default(),
             time_start: win32_app.time_start,
-            last_key_mod: KeyModifiers::default(),
             ime_spot: Vec2::default(),
-            current_cursor: MouseCursor::Default,
             last_mouse_pos: Vec2::default(),
-            fingers_down,
-            ignore_wmsize: 0,
+            pointers_down,
             hwnd: None,
             track_mouse_event: false,
         }
@@ -458,35 +466,43 @@ impl Win32Window {
                 if xcoord < rect.left + EDGE {
                     (*window.win32_app).current_cursor = MouseCursor::Hidden;
                     if ycoord < rect.top + EDGE {
-                        window.do_callback(&mut vec![Event::WindowSetHoverCursor(MouseCursor::NwseResize)]);
+                        window.do_callback(&mut vec![Event::SystemEvent(SystemEvent::WindowSetHoverCursor(
+                            MouseCursor::NwseResize,
+                        ))]);
                         return winuser::HTTOPLEFT;
                     }
                     if ycoord > rect.bottom - EDGE {
-                        window.do_callback(&mut vec![Event::WindowSetHoverCursor(MouseCursor::NeswResize)]);
+                        window.do_callback(&mut vec![Event::SystemEvent(SystemEvent::WindowSetHoverCursor(
+                            MouseCursor::NeswResize,
+                        ))]);
                         return winuser::HTBOTTOMLEFT;
                     }
-                    window.do_callback(&mut vec![Event::WindowSetHoverCursor(MouseCursor::EwResize)]);
+                    window.do_callback(&mut vec![Event::SystemEvent(SystemEvent::WindowSetHoverCursor(MouseCursor::EwResize))]);
                     return winuser::HTLEFT;
                 }
                 if xcoord > rect.right - EDGE {
                     (*window.win32_app).current_cursor = MouseCursor::Hidden;
                     if ycoord < rect.top + EDGE {
-                        window.do_callback(&mut vec![Event::WindowSetHoverCursor(MouseCursor::NeswResize)]);
+                        window.do_callback(&mut vec![Event::SystemEvent(SystemEvent::WindowSetHoverCursor(
+                            MouseCursor::NeswResize,
+                        ))]);
                         return winuser::HTTOPRIGHT;
                     }
                     if ycoord > rect.bottom - EDGE {
-                        window.do_callback(&mut vec![Event::WindowSetHoverCursor(MouseCursor::NwseResize)]);
+                        window.do_callback(&mut vec![Event::SystemEvent(SystemEvent::WindowSetHoverCursor(
+                            MouseCursor::NwseResize,
+                        ))]);
                         return winuser::HTBOTTOMRIGHT;
                     }
-                    window.do_callback(&mut vec![Event::WindowSetHoverCursor(MouseCursor::EwResize)]);
+                    window.do_callback(&mut vec![Event::SystemEvent(SystemEvent::WindowSetHoverCursor(MouseCursor::EwResize))]);
                     return winuser::HTRIGHT;
                 }
                 if ycoord < rect.top + EDGE {
-                    window.do_callback(&mut vec![Event::WindowSetHoverCursor(MouseCursor::NsResize)]);
+                    window.do_callback(&mut vec![Event::SystemEvent(SystemEvent::WindowSetHoverCursor(MouseCursor::NsResize))]);
                     return winuser::HTTOP;
                 }
                 if ycoord > rect.bottom - EDGE {
-                    window.do_callback(&mut vec![Event::WindowSetHoverCursor(MouseCursor::NsResize)]);
+                    window.do_callback(&mut vec![Event::SystemEvent(SystemEvent::WindowSetHoverCursor(MouseCursor::NsResize))]);
                     return winuser::HTBOTTOM;
                 }
                 let mut events = vec![Event::WindowDragQuery(WindowDragQueryEvent {
@@ -499,11 +515,15 @@ impl Win32Window {
                     Event::WindowDragQuery(wd) => match &wd.response {
                         WindowDragQueryResponse::Client => return winuser::HTCLIENT,
                         WindowDragQueryResponse::Caption => {
-                            window.do_callback(&mut vec![Event::WindowSetHoverCursor(MouseCursor::Default)]);
+                            window.do_callback(&mut vec![Event::SystemEvent(SystemEvent::WindowSetHoverCursor(
+                                MouseCursor::Default,
+                            ))]);
                             return winuser::HTCAPTION;
                         }
                         WindowDragQueryResponse::SysMenu => {
-                            window.do_callback(&mut vec![Event::WindowSetHoverCursor(MouseCursor::Default)]);
+                            window.do_callback(&mut vec![Event::SystemEvent(SystemEvent::WindowSetHoverCursor(
+                                MouseCursor::Default,
+                            ))]);
                             return winuser::HTSYSMENU;
                         }
                         _ => (),
@@ -530,11 +550,11 @@ impl Win32Window {
                     };
                     winuser::TrackMouseEvent(&mut tme);
                 }
-                window.send_finger_hover_and_move(window.get_mouse_pos_from_lparam(lparam), Self::get_key_modifiers())
+                window.send_pointer_hover_and_move(window.get_mouse_pos_from_lparam(lparam), Self::get_key_modifiers())
             }
             winuser::WM_MOUSELEAVE => {
                 window.track_mouse_event = false;
-                window.do_callback(&mut vec![Event::FingerHover(FingerHoverEvent {
+                window.do_callback(&mut vec![Event::PointerHover(PointerHoverEvent {
                     digit: 0,
                     window_id: window.window_id,
                     any_down: false,
@@ -550,26 +570,26 @@ impl Win32Window {
             }
             winuser::WM_MOUSEWHEEL => {
                 let delta = (wparam >> 16) as u16 as i16 as f32;
-                window.do_callback(&mut vec![Event::FingerScroll(FingerScrollEvent {
+                window.do_callback(&mut vec![Event::PointerScroll(PointerScrollEvent {
                     digit: 0,
                     window_id: window.window_id,
                     scroll: Vec2 { x: 0.0, y: -delta },
                     abs: window.last_mouse_pos,
                     rel: window.last_mouse_pos,
                     rect: Rect::default(),
-                    input_type: FingerInputType::Mouse,
+                    input_type: PointerInputType::Mouse,
                     modifiers: Self::get_key_modifiers(),
                     handled_x: false,
                     handled_y: false,
                     time: window.time_now(),
                 })]);
             }
-            winuser::WM_LBUTTONDOWN => window.send_finger_down(0, Self::get_key_modifiers()),
-            winuser::WM_LBUTTONUP => window.send_finger_up(0, Self::get_key_modifiers()),
-            winuser::WM_RBUTTONDOWN => window.send_finger_down(1, Self::get_key_modifiers()),
-            winuser::WM_RBUTTONUP => window.send_finger_up(1, Self::get_key_modifiers()),
-            winuser::WM_MBUTTONDOWN => window.send_finger_down(2, Self::get_key_modifiers()),
-            winuser::WM_MBUTTONUP => window.send_finger_up(2, Self::get_key_modifiers()),
+            winuser::WM_LBUTTONDOWN => window.send_pointer_down(0, MouseButton::Left, Self::get_key_modifiers()),
+            winuser::WM_LBUTTONUP => window.send_pointer_up(0, MouseButton::Left, Self::get_key_modifiers()),
+            winuser::WM_RBUTTONDOWN => window.send_pointer_down(1, MouseButton::Right, Self::get_key_modifiers()),
+            winuser::WM_RBUTTONUP => window.send_pointer_up(1, MouseButton::Right, Self::get_key_modifiers()),
+            winuser::WM_MBUTTONDOWN => window.send_pointer_down(2, MouseButton::Other, Self::get_key_modifiers()),
+            winuser::WM_MBUTTONUP => window.send_pointer_up(2, MouseButton::Other, Self::get_key_modifiers()),
             winuser::WM_KEYDOWN | winuser::WM_SYSKEYDOWN => {
                 // detect control/cmd - c / v / x
                 let modifiers = Self::get_key_modifiers();
@@ -605,33 +625,8 @@ impl Win32Window {
                             }
                         }
                         KeyCode::KeyX | KeyCode::KeyC => {
-                            let mut events = vec![Event::TextCopy(TextCopyEvent { response: None })];
+                            let mut events = vec![Event::TextCopy];
                             window.do_callback(&mut events);
-                            match &events[0] {
-                                Event::TextCopy(req) => {
-                                    if let Some(response) = &req.response {
-                                        // plug it into the windows clipboard
-                                        // make utf16 dta
-                                        if winuser::OpenClipboard(ptr::null_mut()) != 0 {
-                                            winuser::EmptyClipboard();
-
-                                            let data: Vec<u16> =
-                                                OsStr::new(response).encode_wide().chain(Some(0).into_iter()).collect();
-
-                                            let h_clipboard_data = winbase::GlobalAlloc(winbase::GMEM_DDESHARE, 2 * data.len());
-
-                                            let h_clipboard_ptr = winbase::GlobalLock(h_clipboard_data) as *mut u16;
-
-                                            std::ptr::copy_nonoverlapping(data.as_ptr(), h_clipboard_ptr, data.len());
-
-                                            winbase::GlobalUnlock(h_clipboard_data);
-                                            winuser::SetClipboardData(winuser::CF_UNICODETEXT, h_clipboard_data);
-                                            winuser::CloseClipboard();
-                                        }
-                                    }
-                                }
-                                _ => (),
-                            };
                         }
                         _ => (),
                     }
@@ -686,22 +681,15 @@ impl Win32Window {
                 // }
             }
             winuser::WM_USER => {
-                let status = if let Ok(_) = (*window.win32_app).race_signals.lock() {
-                    if let Some(status) = (*window.win32_app).usize_to_status.get(&(lparam as usize)) {
-                        Some(*status)
-                    } else {
-                        None
-                    }
-                } else {
-                    None
-                };
-                if let Some(status) = status {
-                    let mut signals = HashMap::new();
-                    let mut set = BTreeSet::new();
-                    set.insert(status);
-                    signals.insert(Signal { signal_id: wparam as usize }, set);
-                    window.do_callback(&mut vec![Event::Signal(SignalEvent { signals })]);
-                }
+                let status_id = usize::from_ne_bytes(lparam.to_ne_bytes());
+                // Panic if we are not 64-bit
+                let status = LocationHash(status_id.try_into().unwrap());
+
+                let mut signals = HashMap::new();
+                let mut set = BTreeSet::new();
+                set.insert(status);
+                signals.insert(Signal { signal_id: wparam as usize }, set);
+                window.do_callback(&mut vec![Event::Signal(SignalEvent { signals })]);
             }
             winuser::WM_CLOSE => {
                 // close requested
@@ -723,7 +711,7 @@ impl Win32Window {
             }
             _ => return winuser::DefWindowProcW(hwnd, msg, wparam, lparam),
         }
-        return 1;
+        1
         // lets get the window
         // Unwinding into foreign code is undefined behavior. So we catch any panics that occur in our
         // code, and if a panic happens we cancel any future operations.
@@ -812,7 +800,7 @@ impl Win32Window {
             if (ex_style & winuser::WS_EX_TOPMOST) != 0 {
                 return true;
             }
-            return false;
+            false
         }
     }
 
@@ -839,7 +827,7 @@ impl Win32Window {
             if wp.showCmd as i32 == winuser::SW_MAXIMIZE {
                 return true;
             }
-            return false;
+            false
         }
     }
 
@@ -926,13 +914,13 @@ impl Win32Window {
 
         self.do_callback(&mut vec![
             Event::WindowGeomChange(WindowGeomChangeEvent { window_id: self.window_id, old_geom, new_geom }),
-            Event::Paint,
+            Event::SystemEvent(SystemEvent::Paint),
         ]);
     }
 
-    pub(crate) fn send_finger_down(&mut self, digit: usize, modifiers: KeyModifiers) {
+    pub(crate) fn send_pointer_down(&mut self, digit: usize, button: MouseButton, modifiers: KeyModifiers) {
         let mut down_count = 0;
-        for is_down in &self.fingers_down {
+        for is_down in &self.pointers_down {
             if *is_down {
                 down_count += 1;
             }
@@ -942,25 +930,26 @@ impl Win32Window {
                 winuser::SetCapture(self.hwnd.unwrap());
             }
         }
-        self.fingers_down[digit] = true;
-        self.do_callback(&mut vec![Event::FingerDown(FingerDownEvent {
+        self.pointers_down[digit] = true;
+        self.do_callback(&mut vec![Event::PointerDown(PointerDownEvent {
             window_id: self.window_id,
             abs: self.last_mouse_pos,
             rel: self.last_mouse_pos,
             rect: Rect::default(),
             digit,
+            button,
             handled: false,
-            input_type: FingerInputType::Mouse,
+            input_type: PointerInputType::Mouse,
             modifiers,
             tap_count: 0,
             time: self.time_now(),
         })]);
     }
 
-    pub(crate) fn send_finger_up(&mut self, digit: usize, modifiers: KeyModifiers) {
-        self.fingers_down[digit] = false;
+    pub(crate) fn send_pointer_up(&mut self, digit: usize, button: MouseButton, modifiers: KeyModifiers) {
+        self.pointers_down[digit] = false;
         let mut down_count = 0;
-        for is_down in &self.fingers_down {
+        for is_down in &self.pointers_down {
             if *is_down {
                 down_count += 1;
             }
@@ -970,7 +959,7 @@ impl Win32Window {
                 winuser::ReleaseCapture();
             }
         }
-        self.do_callback(&mut vec![Event::FingerUp(FingerUpEvent {
+        self.do_callback(&mut vec![Event::PointerUp(PointerUpEvent {
             window_id: self.window_id,
             abs: self.last_mouse_pos,
             rel: self.last_mouse_pos,
@@ -978,19 +967,20 @@ impl Win32Window {
             abs_start: Vec2::default(),
             rel_start: Vec2::default(),
             digit,
+            button,
             is_over: false,
-            input_type: FingerInputType::Mouse,
+            input_type: PointerInputType::Mouse,
             modifiers,
             time: self.time_now(),
         })]);
     }
 
-    pub(crate) fn send_finger_hover_and_move(&mut self, pos: Vec2, modifiers: KeyModifiers) {
+    pub(crate) fn send_pointer_hover_and_move(&mut self, pos: Vec2, modifiers: KeyModifiers) {
         self.last_mouse_pos = pos;
         let mut events = Vec::new();
-        for (digit, down) in self.fingers_down.iter().enumerate() {
+        for (digit, down) in self.pointers_down.iter().enumerate() {
             if *down {
-                events.push(Event::FingerMove(FingerMoveEvent {
+                events.push(Event::PointerMove(PointerMoveEvent {
                     window_id: self.window_id,
                     abs: pos,
                     rel: pos,
@@ -999,13 +989,13 @@ impl Win32Window {
                     abs_start: Vec2::default(),
                     rel_start: Vec2::default(),
                     is_over: false,
-                    input_type: FingerInputType::Mouse,
+                    input_type: PointerInputType::Mouse,
                     modifiers: modifiers.clone(),
                     time: self.time_now(),
                 }));
             }
         }
-        events.push(Event::FingerHover(FingerHoverEvent {
+        events.push(Event::PointerHover(PointerHoverEvent {
             digit: 0,
             window_id: self.window_id,
             abs: pos,
