@@ -2,8 +2,7 @@
 
 #![allow(dead_code)]
 
-use crate::cx::*;
-use libc;
+use crate::*;
 use libc::timeval;
 use std::collections::{BTreeSet, HashMap, VecDeque};
 use std::ffi::CStr;
@@ -81,7 +80,7 @@ pub(crate) struct XlibWindow {
     pub(crate) ime_spot: Vec2,
     pub(crate) current_cursor: MouseCursor,
     pub(crate) last_mouse_pos: Vec2,
-    pub(crate) fingers_down: Vec<bool>,
+    pub(crate) pointers_down: Vec<bool>,
 }
 
 #[derive(Clone)]
@@ -170,6 +169,8 @@ impl XlibApp {
         }
     }
 
+    /// TODO(JP): We learned a lot in improving the OSX event loop and will likely need to apply
+    /// some of the things we did there to here.
     pub(crate) fn event_loop<F>(&mut self, mut event_handler: F)
     where
         F: FnMut(&mut XlibApp, &mut Vec<Event>) -> bool,
@@ -180,7 +181,7 @@ impl XlibApp {
                     as *mut dyn FnMut(&mut XlibApp, &mut Vec<Event>) -> bool,
             );
 
-            self.do_callback(&mut vec![Event::Paint]);
+            self.do_callback(&mut vec![Event::SystemEvent(SystemEvent::Paint)]);
 
             // Record the current time.
             let mut select_time = self.time_now();
@@ -193,18 +194,16 @@ impl XlibApp {
                     // If there are any timers, we set the timeout for select to the `delta_timeout`
                     // of the first timer that should be fired. Otherwise, we set the timeout to
                     // None, so that select will block indefinitely.
-                    let timeout = if let Some(timer) = self.timers.front() {
+                    let timeout = self.timers.front().map(|timer| {
                         // println!("Select wait {}",(timer.delta_timeout.fract() * 1000000.0) as i64);
-                        Some(timeval {
+                        timeval {
                             // `tv_sec` is in seconds, so take the integer part of `delta_timeout`
                             tv_sec: timer.delta_timeout.trunc() as libc::time_t,
                             // `tv_usec` is in microseconds, so take the fractional part of
                             // `delta_timeout` 1000000.0.
                             tv_usec: (timer.delta_timeout.fract() * 1000000.0) as libc::time_t,
-                        })
-                    } else {
-                        None
-                    };
+                        }
+                    });
                     let _nfds = libc::select(
                         self.display_fd + 1,
                         fds.as_mut_ptr(),
@@ -240,7 +239,7 @@ impl XlibApp {
                     self.do_callback(&mut vec![Event::Timer(TimerEvent { timer_id: timer.id })]);
                 }
 
-                while self.display != ptr::null_mut() && X11_sys::XPending(self.display) != 0 {
+                while !self.display.is_null() && X11_sys::XPending(self.display) != 0 {
                     let mut event = mem::MaybeUninit::uninit();
                     X11_sys::XNextEvent(self.display, event.as_mut_ptr());
                     let mut event = event.assume_init();
@@ -292,7 +291,7 @@ impl XlibApp {
                                 );
                                 let ret = ret.assume_init();
                                 //let bytes_after = bytes_after.assume_init();
-                                if ret != ptr::null_mut() && bytes_to_read > 0 {
+                                if !ret.is_null() && bytes_to_read > 0 {
                                     let utf8_slice =
                                         std::slice::from_raw_parts::<u8>(ret as *const _ as *const u8, bytes_to_read as usize);
                                     if let Ok(utf8_string) = String::from_utf8(utf8_slice.to_vec()) {
@@ -379,7 +378,7 @@ impl XlibApp {
                             if crossing.detail == 4 {
                                 if let Some(window_ptr) = self.window_map.get(&crossing.window) {
                                     let window = &mut (**window_ptr);
-                                    window.do_callback(&mut vec![Event::FingerHover(FingerHoverEvent {
+                                    window.do_callback(&mut vec![Event::PointerHover(PointerHoverEvent {
                                         digit: 0,
                                         window_id: window.window_id,
                                         any_down: false,
@@ -429,32 +428,48 @@ impl XlibApp {
                                 window.do_callback(&mut drag_query_events);
                                 // otherwise lets check if we are hover the window edge to resize the window
                                 //println!("{} {}", window.last_window_geom.inner_size.x, pos.x);
-                                window.send_finger_hover_and_move(pos, KeyModifiers::default());
+                                window.send_pointer_hover_and_move(pos, KeyModifiers::default());
                                 let window_size = window.last_window_geom.inner_size;
                                 if pos.x >= 0.0 && pos.x < 10.0 && pos.y >= 0.0 && pos.y < 10.0 {
                                     window.last_nc_mode = Some(_NET_WM_MOVERESIZE_SIZE_TOPLEFT);
-                                    window.do_callback(&mut vec![Event::WindowSetHoverCursor(MouseCursor::NwResize)]);
+                                    window.do_callback(&mut vec![Event::SystemEvent(SystemEvent::WindowSetHoverCursor(
+                                        MouseCursor::NwResize,
+                                    ))]);
                                 } else if pos.x >= 0.0 && pos.x < 10.0 && pos.y >= window_size.y - 10.0 {
                                     window.last_nc_mode = Some(_NET_WM_MOVERESIZE_SIZE_BOTTOMLEFT);
-                                    window.do_callback(&mut vec![Event::WindowSetHoverCursor(MouseCursor::SwResize)]);
+                                    window.do_callback(&mut vec![Event::SystemEvent(SystemEvent::WindowSetHoverCursor(
+                                        MouseCursor::SwResize,
+                                    ))]);
                                 } else if pos.x >= 0.0 && pos.x < 5.0 {
                                     window.last_nc_mode = Some(_NET_WM_MOVERESIZE_SIZE_LEFT);
-                                    window.do_callback(&mut vec![Event::WindowSetHoverCursor(MouseCursor::WResize)]);
+                                    window.do_callback(&mut vec![Event::SystemEvent(SystemEvent::WindowSetHoverCursor(
+                                        MouseCursor::WResize,
+                                    ))]);
                                 } else if pos.x >= window_size.x - 10.0 && pos.y >= 0.0 && pos.y < 10.0 {
                                     window.last_nc_mode = Some(_NET_WM_MOVERESIZE_SIZE_TOPRIGHT);
-                                    window.do_callback(&mut vec![Event::WindowSetHoverCursor(MouseCursor::NeResize)]);
+                                    window.do_callback(&mut vec![Event::SystemEvent(SystemEvent::WindowSetHoverCursor(
+                                        MouseCursor::NeResize,
+                                    ))]);
                                 } else if pos.x >= window_size.x - 10.0 && pos.y >= window_size.y - 10.0 {
                                     window.last_nc_mode = Some(_NET_WM_MOVERESIZE_SIZE_BOTTOMRIGHT);
-                                    window.do_callback(&mut vec![Event::WindowSetHoverCursor(MouseCursor::SeResize)]);
+                                    window.do_callback(&mut vec![Event::SystemEvent(SystemEvent::WindowSetHoverCursor(
+                                        MouseCursor::SeResize,
+                                    ))]);
                                 } else if pos.x >= window_size.x - 5.0 {
                                     window.last_nc_mode = Some(_NET_WM_MOVERESIZE_SIZE_RIGHT);
-                                    window.do_callback(&mut vec![Event::WindowSetHoverCursor(MouseCursor::EResize)]);
+                                    window.do_callback(&mut vec![Event::SystemEvent(SystemEvent::WindowSetHoverCursor(
+                                        MouseCursor::EResize,
+                                    ))]);
                                 } else if pos.y <= 5.0 {
                                     window.last_nc_mode = Some(_NET_WM_MOVERESIZE_SIZE_TOP);
-                                    window.do_callback(&mut vec![Event::WindowSetHoverCursor(MouseCursor::NResize)]);
+                                    window.do_callback(&mut vec![Event::SystemEvent(SystemEvent::WindowSetHoverCursor(
+                                        MouseCursor::NResize,
+                                    ))]);
                                 } else if pos.y > window_size.y - 5.0 {
                                     window.last_nc_mode = Some(_NET_WM_MOVERESIZE_SIZE_BOTTOM);
-                                    window.do_callback(&mut vec![Event::WindowSetHoverCursor(MouseCursor::SResize)]);
+                                    window.do_callback(&mut vec![Event::SystemEvent(SystemEvent::WindowSetHoverCursor(
+                                        MouseCursor::SResize,
+                                    ))]);
                                 } else {
                                     match &drag_query_events[0] {
                                         Event::WindowDragQuery(wd) => match &wd.response {
@@ -488,7 +503,7 @@ impl XlibApp {
                                     self.last_scroll_time = time_now;
                                     // completely arbitrary scroll acceleration curve.
                                     let speed = 1200.0 * (0.2 - 2. * (self.last_scroll_time - last_scroll_time)).max(0.01);
-                                    self.do_callback(&mut vec![Event::FingerScroll(FingerScrollEvent {
+                                    self.do_callback(&mut vec![Event::PointerScroll(PointerScrollEvent {
                                         digit: 0,
                                         window_id: window.window_id,
                                         scroll: Vec2 {
@@ -510,7 +525,7 @@ impl XlibApp {
                                         abs: window.last_mouse_pos,
                                         rel: window.last_mouse_pos,
                                         rect: Rect::default(),
-                                        input_type: FingerInputType::Mouse,
+                                        input_type: PointerInputType::Mouse,
                                         modifiers: self.xkeystate_to_modifiers(button.state),
                                         handled_x: false,
                                         handled_y: false,
@@ -559,7 +574,8 @@ impl XlibApp {
                                             );
                                         }
                                     } else {
-                                        window.send_finger_down(button.button as usize, self.xkeystate_to_modifiers(button.state))
+                                        window
+                                            .send_pointer_down(button.button as usize, self.xkeystate_to_modifiers(button.state))
                                     }
                                 }
                             }
@@ -571,7 +587,7 @@ impl XlibApp {
                             let button = event.xbutton;
                             if let Some(window_ptr) = self.window_map.get(&button.window) {
                                 let window = &mut (**window_ptr);
-                                window.send_finger_up(button.button as usize, self.xkeystate_to_modifiers(button.state))
+                                window.send_pointer_up(button.button as usize, self.xkeystate_to_modifiers(button.state))
                             }
                         }
                         X11_sys::KeyPress => {
@@ -605,26 +621,8 @@ impl XlibApp {
                                                 */
                                             }
                                             KeyCode::KeyX | KeyCode::KeyC => {
-                                                let mut events = vec![Event::TextCopy(TextCopyEvent { response: None })];
+                                                let mut events = vec![Event::TextCopy];
                                                 self.do_callback(&mut events);
-                                                match &events[0] {
-                                                    Event::TextCopy(req) => {
-                                                        if let Some(response) = &req.response {
-                                                            // store the text on the clipboard
-                                                            self.clipboard = response.clone();
-                                                            // lets set the owner
-                                                            println!("Set selection owner");
-                                                            X11_sys::XSetSelectionOwner(
-                                                                self.display,
-                                                                self.atom_clipboard,
-                                                                window.window.unwrap(),
-                                                                event.xkey.time,
-                                                            );
-                                                            X11_sys::XFlush(self.display);
-                                                        }
-                                                    }
-                                                    _ => (),
-                                                };
                                             }
                                             _ => (),
                                         }
@@ -716,11 +714,11 @@ impl XlibApp {
                 } else {
                     Vec::new()
                 };
-                if proc_signals.len() > 0 {
+                if !proc_signals.is_empty() {
                     self.do_callback(&mut proc_signals);
                 }
 
-                self.do_callback(&mut vec![Event::Paint]);
+                self.do_callback(&mut vec![Event::SystemEvent(SystemEvent::Paint)]);
             }
 
             self.event_callback = None;
@@ -842,7 +840,7 @@ impl XlibApp {
                 }
             }
         }
-        return None;
+        None
     }
 
     pub(crate) fn set_mouse_cursor(&mut self, cursor: MouseCursor) {
@@ -880,7 +878,7 @@ impl XlibApp {
             if let Some(x11_cursor) = x11_cursor {
                 unsafe {
                     for (k, v) in &self.window_map {
-                        if !(**v).window.is_none() {
+                        if (**v).window.is_some() {
                             X11_sys::XDefineCursor(self.display, *k, x11_cursor);
                         }
                     }
@@ -1040,12 +1038,38 @@ impl XlibApp {
             _ => KeyCode::Unknown,
         }
     }
+
+    pub(crate) fn copy_text_to_clipboard(_text: &str) {
+        // store the text on the clipboard
+        // TODO(JP): do this in a way that is not window-specific;
+        // see e.g. https://stackoverflow.com/a/41970363
+        // let app = *GLOBAL_XLIB_APP;
+        // app.clipboard = text.clone();
+        // X11_sys::XSetSelectionOwner(
+        //     app.display,
+        //     app.atom_clipboard,
+        //     window.window.unwrap(),
+        //     event.xkey.time,
+        // );
+        // X11_sys::XFlush(app.display);
+    }
+}
+
+/// In x11, 1 is left, 3 is right, and 2 is middle click.
+/// Refer: <http://xahlee.info/linux/linux_x11_mouse_button_number.html>
+/// Double checked this using `xev` command in ubuntu.
+fn get_mouse_button_from_digit(digit: usize) -> MouseButton {
+    match digit {
+        1 => MouseButton::Left,
+        3 => MouseButton::Right,
+        _ => MouseButton::Other,
+    }
 }
 
 impl XlibWindow {
     pub(crate) fn new(xlib_app: &mut XlibApp, window_id: usize) -> XlibWindow {
-        let mut fingers_down = Vec::new();
-        fingers_down.resize(NUM_FINGERS, false);
+        let mut pointers_down = Vec::new();
+        pointers_down.resize(NUM_POINTERS, false);
 
         XlibWindow {
             window: None,
@@ -1061,7 +1085,7 @@ impl XlibWindow {
             ime_spot: Vec2::default(),
             current_cursor: MouseCursor::Default,
             last_mouse_pos: Vec2::default(),
-            fingers_down,
+            pointers_down,
         }
     }
 
@@ -1094,12 +1118,13 @@ impl XlibWindow {
                 | X11_sys::LeaveWindowMask) as c_long;
 
             let dpi_factor = self.get_dpi_factor();
+            let window_position = position.unwrap_or(Vec2 { x: 150.0, y: 60.0 });
             // Create a window
             let window = X11_sys::XCreateWindow(
                 display,
                 root_window,
-                if position.is_some() { position.unwrap().x } else { 150.0 } as i32,
-                if position.is_some() { position.unwrap().y } else { 60.0 } as i32,
+                window_position.x as i32,
+                window_position.y as i32,
                 (size.x * dpi_factor) as u32,
                 (size.y * dpi_factor) as u32,
                 0,
@@ -1235,7 +1260,7 @@ impl XlibWindow {
 
             self.child_windows.push(XlibChildWindow { window: new_child, x, y, w, h, visible: true });
 
-            return Some(new_child);
+            Some(new_child)
         }
     }
 
@@ -1363,7 +1388,7 @@ impl XlibWindow {
             let n_item = n_item.assume_init();
             //let bytes_after = bytes_after.assume_init();
             let properties = properties.assume_init();
-            if result == 0 && properties != ptr::null_mut() {
+            if result == 0 && !properties.is_null() {
                 let items = std::slice::from_raw_parts::<c_ulong>(properties as *mut _, n_item as usize);
                 for item in items {
                     if *item == xlib_app.atom_new_wm_state_maximized_horz || *item == xlib_app.atom_new_wm_state_maximized_vert {
@@ -1392,7 +1417,7 @@ impl XlibWindow {
             let display = (*self.xlib_app).display;
             X11_sys::XGetWindowAttributes(display, self.window.unwrap(), xwa.as_mut_ptr());
             let xwa = xwa.assume_init();
-            return Vec2 { x: xwa.x as f32, y: xwa.y as f32 };
+            Vec2 { x: xwa.x as f32, y: xwa.y as f32 }
             /*
             let mut child = mem::uninitialized();
             let default_screen = X11_sys::XDefaultScreen(display);
@@ -1411,7 +1436,7 @@ impl XlibWindow {
             let display = (*self.xlib_app).display;
             X11_sys::XGetWindowAttributes(display, self.window.unwrap(), xwa.as_mut_ptr());
             let xwa = xwa.assume_init();
-            return Vec2 { x: xwa.width as f32 / dpi_factor, y: xwa.height as f32 / dpi_factor };
+            Vec2 { x: xwa.width as f32 / dpi_factor, y: xwa.height as f32 / dpi_factor }
         }
     }
 
@@ -1421,7 +1446,7 @@ impl XlibWindow {
             let display = (*self.xlib_app).display;
             X11_sys::XGetWindowAttributes(display, self.window.unwrap(), xwa.as_mut_ptr());
             let xwa = xwa.assume_init();
-            return Vec2 { x: xwa.width as f32, y: xwa.height as f32 };
+            Vec2 { x: xwa.width as f32, y: xwa.height as f32 }
         }
     }
 
@@ -1436,7 +1461,7 @@ impl XlibWindow {
             //return 2.0;
             let display = (*self.xlib_app).display;
             let resource_string = X11_sys::XResourceManagerString(display);
-            if resource_string == std::ptr::null_mut() {
+            if resource_string.is_null() {
                 return 1.0;
             }
             let db = X11_sys::XrmGetStringDatabase(resource_string);
@@ -1451,11 +1476,11 @@ impl XlibWindow {
             );
             //let ty = ty.assume_init();
             let value = value.assume_init();
-            if value.addr == std::ptr::null_mut() {
-                return 1.0; // TODO find some other way to figure it out
+            if value.addr.is_null() {
+                1.0 // TODO find some other way to figure it out
             } else {
                 let dpi: f32 = CStr::from_ptr(value.addr).to_str().unwrap().parse().unwrap();
-                return dpi / 96.0;
+                dpi / 96.0
             }
         }
     }
@@ -1478,7 +1503,7 @@ impl XlibWindow {
 
         self.do_callback(&mut vec![
             Event::WindowGeomChange(WindowGeomChangeEvent { window_id: self.window_id, old_geom, new_geom }),
-            Event::Paint,
+            Event::SystemEvent(SystemEvent::Paint),
         ]);
     }
 
@@ -1490,9 +1515,9 @@ impl XlibWindow {
         self.do_callback(&mut vec![Event::AppFocusLost]);
     }
 
-    pub(crate) fn send_finger_down(&mut self, digit: usize, modifiers: KeyModifiers) {
+    pub(crate) fn send_pointer_down(&mut self, digit: usize, modifiers: KeyModifiers) {
         let mut down_count = 0;
-        for is_down in &self.fingers_down {
+        for is_down in &self.pointers_down {
             if *is_down {
                 down_count += 1;
             }
@@ -1500,25 +1525,26 @@ impl XlibWindow {
         if down_count == 0 {
             //unsafe {winuser::SetCapture(self.hwnd.unwrap());}
         }
-        self.fingers_down[digit] = true;
-        self.do_callback(&mut vec![Event::FingerDown(FingerDownEvent {
+        self.pointers_down[digit] = true;
+        self.do_callback(&mut vec![Event::PointerDown(PointerDownEvent {
             window_id: self.window_id,
             abs: self.last_mouse_pos,
             rel: self.last_mouse_pos,
             rect: Rect::default(),
             digit,
+            button: get_mouse_button_from_digit(digit),
             handled: false,
-            input_type: FingerInputType::Mouse,
+            input_type: PointerInputType::Mouse,
             modifiers,
             tap_count: 0,
             time: self.time_now(),
         })]);
     }
 
-    pub(crate) fn send_finger_up(&mut self, digit: usize, modifiers: KeyModifiers) {
-        self.fingers_down[digit] = false;
+    pub(crate) fn send_pointer_up(&mut self, digit: usize, modifiers: KeyModifiers) {
+        self.pointers_down[digit] = false;
         let mut down_count = 0;
-        for is_down in &self.fingers_down {
+        for is_down in &self.pointers_down {
             if *is_down {
                 down_count += 1;
             }
@@ -1526,7 +1552,7 @@ impl XlibWindow {
         if down_count == 0 {
             // unsafe {winuser::ReleaseCapture();}
         }
-        self.do_callback(&mut vec![Event::FingerUp(FingerUpEvent {
+        self.do_callback(&mut vec![Event::PointerUp(PointerUpEvent {
             window_id: self.window_id,
             abs: self.last_mouse_pos,
             rel: self.last_mouse_pos,
@@ -1534,19 +1560,20 @@ impl XlibWindow {
             abs_start: Vec2::default(),
             rel_start: Vec2::default(),
             digit,
+            button: get_mouse_button_from_digit(digit),
             is_over: false,
-            input_type: FingerInputType::Mouse,
+            input_type: PointerInputType::Mouse,
             modifiers,
             time: self.time_now(),
         })]);
     }
 
-    pub(crate) fn send_finger_hover_and_move(&mut self, pos: Vec2, modifiers: KeyModifiers) {
+    pub(crate) fn send_pointer_hover_and_move(&mut self, pos: Vec2, modifiers: KeyModifiers) {
         self.last_mouse_pos = pos;
         let mut events = Vec::new();
-        for (digit, down) in self.fingers_down.iter().enumerate() {
+        for (digit, down) in self.pointers_down.iter().enumerate() {
             if *down {
-                events.push(Event::FingerMove(FingerMoveEvent {
+                events.push(Event::PointerMove(PointerMoveEvent {
                     window_id: self.window_id,
                     abs: pos,
                     rel: pos,
@@ -1555,13 +1582,13 @@ impl XlibWindow {
                     abs_start: Vec2::default(),
                     rel_start: Vec2::default(),
                     is_over: false,
-                    input_type: FingerInputType::Mouse,
+                    input_type: PointerInputType::Mouse,
                     modifiers: modifiers.clone(),
                     time: self.time_now(),
                 }));
             }
         }
-        events.push(Event::FingerHover(FingerHoverEvent {
+        events.push(Event::PointerHover(PointerHoverEvent {
             digit: 0,
             window_id: self.window_id,
             abs: pos,
