@@ -12,18 +12,13 @@
 //! [`DrawCall`]s contain the actual data that needs to be drawn to the screen.
 //!
 //! [`View`]s are groups of draw calls, have some special features, such as
-//! scrolling/clipping and caching.
+//! scrolling/clipping.
 //!
 //! Note that one level higher, we have a hierarchy of [`Pass`]es.
 
-use crate::cx::*;
+use crate::*;
 
-/// Gets returned by [`View::begin_view`].
-///
-/// TODO(JP): maybe better to make this an enum with descriptive names?
-/// It looks like this is done as a [`Result`] so you can use the [`?`] shorthand,
-/// but not sure if that is really clear enough?
-pub type ViewRedraw = Result<(), ()>;
+use crate::Debugger;
 
 /// A group of [`DrawCall`]s.
 ///
@@ -31,10 +26,15 @@ pub type ViewRedraw = Result<(), ()>;
 /// so you can find more information there.
 ///
 /// A [`View`] has a few special features:
-/// * It has its own
+/// * It has its own [`Rect`], within which everything is clipped (see also [`DrawUniforms`]).
+///   This typically gets set by the return value of [`Cx::end_typed_box`] for the [`CxLayoutBox`] that
+///   is associated with the [`View`]. TODO(JP): Look into decoupling [`CxLayoutBox`] from [`View`].
+/// * It can scroll (but does not have to; again see also [`DrawUniforms`]).
+/// * It has its own set of [`DrawCall`]s, which are isolated from the [`DrawCall`]s of the
+///   parent [`View`].
 ///
 /// See also [`ViewArea`], which is an [`Area`] pointer to a [`View`].
-#[derive(Clone)]
+#[derive(Clone, Default)]
 pub struct View {
     /// The index of the corresponding [`CxView`] in [`Cx::views`].
     pub view_id: Option<usize>,
@@ -43,61 +43,42 @@ pub struct View {
     /// Whether this [`View`] is an overlay/popup, which means all [`DrawCall`]s underneath it
     /// will get rendered last.
     pub(crate) is_overlay: bool,
-    /// Whether this [`View`] will always redraw. Currently only used internally in fonts.
-    ///
-    /// TODO(JP): We can probably do this some other way? Like just explicitly mark it for redraw
-    /// before the [`View::begin_view`] or so?
-    pub(crate) always_redraw: bool,
+
+    debugger: Debugger,
 }
 
 impl View {
     /// Creates a new, empty [`View`].
-    pub fn new() -> Self {
-        Self { redraw_id: 0, is_overlay: false, always_redraw: false, view_id: None }
-    }
+
     /// See [`View::is_overlay`].
+    #[must_use]
     pub fn with_is_overlay(self, is_overlay: bool) -> Self {
         Self { is_overlay, ..self }
     }
-    /// See [`View::always_redraw`].
-    pub(crate) fn with_always_redraw(self, always_redraw: bool) -> Self {
-        Self { always_redraw, ..self }
-    }
 
-    /// Register the [`View`] in the draw tree. Returns [`Result::Err`] if rendering
-    /// should be skipped since it wasn't marked for redrawing. Returns [`Result::Ok`]
-    /// when drawing should happen, ie. when it was marked for redrawing, or when
-    /// drawing the [`View`] for the first time (when it doesn't have a
-    /// [`View::view_id`] yet).
+    /// Register the [`View`] in the draw tree.
     ///
-    /// This also creates a new [`Turtle`] with the [`Layout`] that you pass in.
+    /// This also creates a new [`CxLayoutBox`] with the [`LayoutSize`] that is passed in.
     /// Note that you should not create a [`View`] just in order to get a new
-    /// [`Turtle`], since creating a [`View`] is relatively expensive -- no
+    /// [`CxLayoutBox`], since creating a [`View`] is relatively expensive -- no
     /// [`DrawCall`]s inside this [`View`] will get merged with ones outside of
     /// it, so adding too many [`View`]s will create too many individual calls to
     /// the GPU.
     ///
-    /// TODO(JP): Perhaps we should decouple [`Turtle`] and [`View`] altogether?
-    ///
-    /// TODO(JP): If you move where a [`View`] gets rendered in the hierarchy,
-    /// then you have to be really sure that you mark the [`View`] that it gets
-    /// moved from as needing a redraw (e.g. using [`View::redraw_view`]), otherwise
-    /// the [`CxView`] might end up with two parents!
-    pub fn begin_view(&mut self, cx: &mut Cx, layout: Layout) -> ViewRedraw {
+    /// TODO(JP): Perhaps we should decouple [`CxLayoutBox`] and [`View`] altogether?
+    pub fn begin_view(&mut self, cx: &mut Cx, layout_size: LayoutSize) {
+        self.begin_view_with_layout(cx, Layout { direction: Direction::Down, layout_size, ..Layout::default() });
+    }
+
+    fn begin_view_with_layout(&mut self, cx: &mut Cx, layout: Layout) {
         if !cx.in_redraw_cycle {
             panic!("calling begin_view outside of redraw cycle is not possible!");
         }
+        assert!(cx.shader_group_instance_offsets.is_empty(), "Can't use begin_view inside a shader group");
 
         // check if we have a pass id parent
         let pass_id = *cx.pass_stack.last().expect("No pass found when begin_view");
 
-        let nesting_view_id = if !cx.view_stack.is_empty() {
-            *cx.view_stack.last().unwrap()
-        } else {
-            // return the root draw list
-            0
-        };
-
         if self.view_id.is_none() {
             // we need a draw_list_id
             self.view_id = Some(cx.views.len());
@@ -105,8 +86,6 @@ impl View {
             let cxview = &mut cx.views[self.view_id.unwrap()];
             cxview.redraw_id = cx.redraw_id;
             cxview.pass_id = pass_id;
-            // set nesting draw list id for incremental repaint scanning
-            cx.views[self.view_id.unwrap()].nesting_view_id = nesting_view_id;
         }
 
         let view_id = self.view_id.unwrap();
@@ -115,8 +94,8 @@ impl View {
             // we are the first view on a window
             let cxpass = &mut cx.passes[pass_id];
             cxpass.main_view_id = Some(view_id);
-            // we should take the window geometry and abs position as our turtle layout
-            (Layout { abs_origin: Some(Vec2 { x: 0., y: 0. }), abs_size: Some(cxpass.pass_size), ..layout }, true)
+            // we should take the window geometry and abs position as our box layout
+            (Layout { absolute: true, abs_size: Some(cxpass.pass_size), ..layout }, true)
         } else {
             (layout, false)
         };
@@ -165,13 +144,14 @@ impl View {
             }
         }
 
-        if !self.always_redraw && cx.views[view_id].draw_calls_len != 0 && !cx.view_will_redraw(view_id) {
-            // walk the turtle because we aren't drawing
-            let w = Width::Fix(cx.views[view_id].rect.size.x);
-            let h = Height::Fix(cx.views[view_id].rect.size.y);
-            cx.walk_turtle(Walk { width: w, height: h, margin: override_layout.walk.margin });
-            return Err(());
-        }
+        // TODO(JP): Do we want to keep this? We don't really use this for anything except as a
+        // convenience. I talked with Rik about redrawing of [`View`]s, and one idea was to always
+        // fully invalidate the closest [`View`] parent that did not have a [`Layout`] with
+        // [`Width::Compute`] or [`Height::Compute`], but that seems to fragile to me. It would be
+        // better to check if a [`CxView::rect`] actually changed and in that case trigger a redraw
+        // or even a panic (with some way of manually overriding the panic). So anyway, I think we
+        // should strive to remove this after all.
+        cx.begin_typed_box(CxBoxType::View, override_layout);
 
         // prepare drawlist for drawing
         let cxview = &mut cx.views[view_id];
@@ -181,27 +161,34 @@ impl View {
         assert_eq!(cxview.pass_id, pass_id);
 
         // update drawlist ids
-        let last_redraw_id = cxview.redraw_id;
         self.redraw_id = cx.redraw_id;
         cxview.redraw_id = cx.redraw_id;
         cxview.draw_calls_len = 0;
 
         cx.view_stack.push(view_id);
 
-        let old_area = Area::View(ViewArea { view_id, redraw_id: last_redraw_id });
-        let new_area = Area::View(ViewArea { view_id, redraw_id: cx.redraw_id });
-        cx.update_area_refs(old_area, new_area);
-
-        cx.begin_turtle(override_layout, new_area);
-
         if is_root_for_pass {
             cx.passes[pass_id].paint_dirty = true;
         }
-
-        Ok(())
     }
 
-    /// End the [`View`], by ending the [`Turtle`]. Returns a [`ViewArea`] that
+    fn is_main_view(view_id: usize, cx: &mut Cx) -> bool {
+        if let Some(window_id) = cx.window_stack.last() {
+            if let Some(main_pass_id) = cx.windows[*window_id].main_pass_id {
+                let pass_id = cx.views[view_id].pass_id;
+                let cxpass = &cx.passes[pass_id];
+                if let Some(main_view_id) = cxpass.main_view_id {
+                    if main_view_id == view_id && main_pass_id == pass_id {
+                        // we are the main view of a main pass of a window
+                        return true;
+                    }
+                }
+            }
+        }
+        false
+    }
+
+    /// End the [`View`], by ending the [`CxLayoutBox`]. Returns a [`ViewArea`] that
     /// you can hold onto.
     ///
     /// Should only be called if [`View::begin_view`] returned [`Result::Ok`].
@@ -209,27 +196,25 @@ impl View {
     /// TODO(JP): Is the [`ViewArea`] redundant, since it basically contains the
     /// same information as the [`View`] itself?
     pub fn end_view(&mut self, cx: &mut Cx) -> Area {
+        assert!(cx.shader_group_instance_offsets.is_empty(), "Can't use end_view inside a shader group");
+
         let view_id = self.view_id.unwrap();
+
+        if cx.debug_flags.enable_layout_debugger && View::is_main_view(view_id, cx) {
+            self.debugger.draw(cx);
+        }
+
         let view_area = Area::View(ViewArea { view_id, redraw_id: cx.redraw_id });
-        let rect = cx.end_turtle(view_area);
-        let cxview = &mut cx.views[view_id];
-        cxview.rect = rect;
+        // Make sure that ViewArea would also be aligned when underlying calls getting moved
+        cx.layout_box_align_list.push(view_area);
+
+        let rect = cx.end_typed_box(CxBoxType::View);
+        cx.views[view_id].rect = rect;
         cx.view_stack.pop();
         view_area
     }
 
-    /// Whether or not the [`View`] will redraw.
-    ///
-    /// See [`Cx::view_will_redraw`].
-    pub fn view_will_redraw(&self, cx: &mut Cx) -> bool {
-        if let Some(view_id) = self.view_id {
-            cx.view_will_redraw(view_id)
-        } else {
-            true
-        }
-    }
-
-    /// Get the [`Rect`] that the [`Turtle`] associated with the [`View`]
+    /// Get the [`Rect`] that the [`CxLayoutBox`] associated with the [`View`]
     /// returned.
     ///
     /// TODO(JP): Should we return an [`Option<Rect>`] instead of just
@@ -246,34 +231,6 @@ impl View {
         Rect::default()
     }
 
-    /// Print debug information about this [`View`].
-    pub fn set_view_debug(&self, cx: &mut Cx, view_debug: CxViewDebug) {
-        if let Some(view_id) = self.view_id {
-            let cxview = &mut cx.views[view_id];
-            cxview.debug = Some(view_debug);
-        }
-    }
-
-    /// Redraws the [`View`] using [`Cx::redraw_child_area`].
-    ///
-    /// If the [`View`] hasn't been instantiated yet, we're conservative and
-    /// just redraw everything using [`Area::All`].
-    ///
-    /// TODO(JP): Should we drop this function and just have users call
-    /// `Cx::redraw_child_area(view.area())`? That won't do the [`Area::All`]
-    /// thing mentioned above, but maybe that is bad anyway? Or maybe it's good,
-    /// and we should just change `[Cx::redraw_child_area`] to redraw everything
-    /// on [`Area::Empty`] (and get rid of [`Area::All`] completely)?
-    pub fn redraw_view(&self, cx: &mut Cx) {
-        if let Some(view_id) = self.view_id {
-            let cxview = &cx.views[view_id];
-            let area = Area::View(ViewArea { view_id, redraw_id: cxview.redraw_id });
-            cx.redraw_child_area(area);
-        } else {
-            cx.redraw_child_area(Area::All)
-        }
-    }
-
     /// Returns an [`Area::View`] for this [`View`], or [`Area::Empty`] if the
     /// [`View`] hasn't been instantiated yet.
     pub fn area(&self) -> Area {
@@ -297,26 +254,8 @@ impl View {
 }
 
 impl Cx {
-    /// Create and returns a new [`DrawCall`] for the given [`Shader`], even if
-    /// one already exists.
-    ///
-    /// Will be nested in the current [`CxView`].
-    ///
-    /// You don't have to use the [`DrawCall`] that gets returned. Instead you
-    /// may just call this to force a new [`DrawCall`] to be created for the
-    /// [`Shader`], and then use other functions to append to it.
-    pub fn new_draw_call(&mut self, shader: Shader) -> &mut DrawCall {
-        self.get_draw_call(false, shader, None)
-    }
-
     /// Returns an existing [`DrawCall`] based on the given [`Shader`], or
     /// creates a new one if none can be found in the current [`CxView`].
-    fn append_to_draw_call(&mut self, shader: Shader, slots: usize) -> &mut DrawCall {
-        self.get_draw_call(true, shader, Some(slots))
-    }
-
-    /// Common code between [`Cx::new_draw_call`] and [`Cx::append_to_draw_call`].
-    /// Either creates a new [`DrawCall`] or appends to an existing one.
     ///
     /// Reuses an existing [`DrawCall`] if [`CxView::draw_calls_len`] is less than
     /// [`CxView::draw_calls.len()`], so we can reuse existing GPU resources.
@@ -324,39 +263,52 @@ impl Cx {
     /// TODO(JP): It's unclear to me if the reusing of GPU resources in this way
     /// is beneficial. And if it is, if it should instead be done in the
     /// platform-specific code instead.
-    fn get_draw_call(&mut self, append: bool, shader: Shader, slots: Option<usize>) -> &mut DrawCall {
-        let sh = &self.shaders[shader.shader_id];
+    fn create_draw_call(&mut self, shader_id: usize, props: DrawCallProps) -> &mut DrawCall {
+        assert!(self.in_redraw_cycle, "Must be in redraw cycle to append to draw calls");
+
+        let sh = &self.shaders[shader_id];
 
         let current_view_id = *self.view_stack.last().unwrap();
         let cxview = &mut self.views[current_view_id];
         let draw_call_id = cxview.draw_calls_len;
 
-        if append {
-            if let Some(index) = cxview.find_appendable_drawcall(shader) {
-                return &mut cxview.draw_calls[index];
+        // Find a draw call to append to.
+        if props.is_batchable() {
+            let shader_group_size = self.shader_group_instance_offsets.len();
+            if shader_group_size > 0 {
+                // If we're in a shader group then the given shader must be part of the group, so we just
+                // search within the group.
+                assert!(cxview.draw_calls_len >= shader_group_size);
+                for index in cxview.draw_calls_len - shader_group_size..cxview.draw_calls_len {
+                    if cxview.draw_calls[index].shader_id == shader_id {
+                        return &mut cxview.draw_calls[index];
+                    }
+                }
+                panic!("Trying to use Shader within a shader group that isn't part of the group");
+            } else {
+                // Do the most basic of [`DrawCall`] batching, by checking if the very last [`DrawCall`] matches
+                // the shader that we're drawing, and if so, appending to that.
+                if cxview.draw_calls_len > 0 && !self.debug_flags.disable_draw_call_batching {
+                    let dc = &mut cxview.draw_calls[cxview.draw_calls_len - 1];
+                    if dc.props.is_batchable() && dc.sub_view_id == 0 && dc.shader_id == shader_id {
+                        return &mut cxview.draw_calls[cxview.draw_calls_len - 1];
+                    }
+                }
             }
         }
 
         // add one
         cxview.draw_calls_len += 1;
 
-        if let Some(slots) = slots {
-            if slots != sh.mapping.instance_props.total_slots {
-                log!("Warning, instance misalignment between struct and shader in {}", sh.name)
-            }
-        }
         // see if we need to add a new one
         if draw_call_id >= cxview.draw_calls.len() {
             cxview.draw_calls.push(DrawCall {
-                geometry: None,
+                props,
                 draw_call_id,
                 view_id: current_view_id,
                 redraw_id: self.redraw_id,
-                scroll_sticky_horizontal: false,
-                scroll_sticky_vertical: false,
-                in_many_instances: false,
                 sub_view_id: 0,
-                shader,
+                shader_id,
                 instances: Vec::new(),
                 draw_uniforms: DrawUniforms::default(),
                 user_uniforms: {
@@ -379,8 +331,8 @@ impl Cx {
         }
         // reuse an older one, keeping all GPU resources attached
         let dc = &mut cxview.draw_calls[draw_call_id];
-        dc.shader = shader;
-        dc.geometry = None;
+        dc.shader_id = shader_id;
+        dc.props = props;
         dc.sub_view_id = 0; // make sure its recognised as a draw call
                             // truncate buffers and set update frame
         dc.redraw_id = self.redraw_id;
@@ -391,85 +343,147 @@ impl Cx {
         dc.textures_2d.resize(sh.mapping.textures.len(), 0);
         dc.instance_dirty = true;
         dc.uniforms_dirty = true;
-        dc.scroll_sticky_horizontal = false;
-        dc.scroll_sticky_vertical = false;
         dc
     }
 
-    /// Start manually editing [`DrawCall::instances`], through a [`ManyInstances`]
-    /// temporary object.
-    ///
-    /// TODO(JP): See [`ManyInstances`] for a TODO on whether we even need
-    /// [`ManyInstances`].
-    pub fn begin_many_instances(&mut self, shader: Shader, slots: usize) -> ManyInstances {
-        let dc = self.append_to_draw_call(shader, slots);
-        let mut instances = Vec::new();
-        if dc.in_many_instances {
-            panic!("please call end_many_instances before calling begin_many_instances again")
-        }
-        dc.in_many_instances = true;
-        std::mem::swap(&mut instances, &mut dc.instances);
-        ManyInstances {
-            instance_area: InstanceRangeArea {
-                view_id: dc.view_id,
-                draw_call_id: dc.draw_call_id,
-                instance_count: 0,
-                instance_offset: instances.len(),
-                redraw_id: dc.redraw_id,
-            },
-            instances,
-        }
-    }
-
-    /// End manual editing of [`DrawCall::instances`]. Returns an [`Area::InstanceRange`]
-    /// that points to the range of instance you just added.
-    ///
-    /// TODO(JP): See [`ManyInstances`] for a TODO on whether we even need
-    /// [`ManyInstances`].
-    ///
-    /// TODO(JP): Removing or editing instances doesn't seem to be supported; those won't
-    /// be described in the [`Area`] (in case of removing you might even get a crash
-    /// since [`InstanceRangeArea::instance_count` is unsigned]).
-    pub fn end_many_instances(&mut self, mut many_instances: ManyInstances) -> Area {
-        let mut ia = many_instances.instance_area;
-        let cxview = &mut self.views[ia.view_id];
-        let dc = &mut cxview.draw_calls[ia.draw_call_id];
-
-        if !dc.in_many_instances {
-            panic!("please call begin_many_instances before calling end_many_instances")
-        }
-        dc.in_many_instances = false;
-        std::mem::swap(&mut many_instances.instances, &mut dc.instances);
-        let sh = &self.shaders[dc.shader.shader_id];
-        ia.instance_count = (dc.instances.len() - ia.instance_offset) / sh.mapping.instance_props.total_slots;
-        Area::InstanceRange(ia)
-    }
-
-    /// Add a single instance to [`DrawCall::instances`].
-    ///
-    /// Uses [`Cx::append_to_draw_call`] under the hood to find the [`DrawCall`]
-    /// to add to.
-    ///
-    /// TODO(JP): This is convenient, but might be a bit too much indirection for
-    /// my taste. Why not have this function on [`DrawCall`] directly, and have
-    /// users call [`Cx::append_to_draw_call`] directly?
-    pub fn add_instance(&mut self, shader: Shader, data: &[f32]) -> Area {
-        let total_instance_slots = self.shaders[shader.shader_id].mapping.instance_props.total_slots;
-        let dc = self.append_to_draw_call(shader, data.len());
-        let instance_count = data.len() / total_instance_slots;
-        let check = data.len() % total_instance_slots;
-        if check > 0 {
-            panic!("Data not multiple of total slots");
+    /// Add a slice of instances to [`DrawCall::instances`]. See [`Cx::add_instances`].
+    fn add_instances_internal<T: Sized>(&mut self, shader: &'static Shader, data: &[T], props: DrawCallProps) -> Area {
+        if data.is_empty() {
+            // This is important, because otherwise you can call this function with empty data in order to force
+            // a particular ordering of `DrawCall`s, and then depend on batching of `DrawCall`s. That should
+            // generally be avoided -- we might change how `DrawCall` batching works in the future.
+            return Area::Empty;
         }
+        let shader_id = self.get_shader_id(shader);
+        let total_instance_slots = self.shaders[shader_id].mapping.instance_props.total_slots;
+        assert_eq!(total_instance_slots * std::mem::size_of::<f32>(), std::mem::size_of::<T>());
+        let dc = self.create_draw_call(shader_id, props);
         let ia = InstanceRangeArea {
             view_id: dc.view_id,
             draw_call_id: dc.draw_call_id,
-            instance_count,
+            instance_count: data.len(),
             instance_offset: dc.instances.len(),
             redraw_id: dc.redraw_id,
         };
-        dc.instances.extend_from_slice(data);
-        Area::InstanceRange(ia)
+        dc.instances.extend_from_slice(data.as_f32_slice());
+        let area = Area::InstanceRange(ia);
+        self.add_to_box_align_list(area);
+        area
+    }
+
+    /// Add a slice of instances to [`DrawCall::instances`].
+    ///
+    /// Supports appending any data that has the correct size.
+    ///
+    /// You should assume that any call to this creates a new [`DrawCall`], even though in reality we might
+    /// batch certain [`DrawCall`]s. For more information about [`DrawCall`] batching, see [`Shader`].
+    ///
+    /// Uses [`Cx::create_draw_call`] under the hood to find the [`DrawCall`]
+    /// to add to.
+    pub fn add_instances<T: Sized>(&mut self, shader: &'static Shader, data: &[T]) -> Area {
+        assert!(shader.build_geom.is_some(), "Can't add instances without `build_geom` defined");
+
+        self.add_instances_internal(shader, data, DrawCallProps::default())
+    }
+
+    /// Add a slice of instances while specifying a custom Geometry
+    pub fn add_mesh_instances<T: Sized>(&mut self, shader: &'static Shader, data: &[T], gpu_geometry: GpuGeometry) -> Area {
+        assert!(self.shader_group_instance_offsets.is_empty(), "Can't add mesh instances when in a shader group");
+
+        self.add_instances_internal(shader, data, DrawCallProps { gpu_geometry: Some(gpu_geometry), ..Default::default() })
+    }
+
+    /// By default, [`DrawCall`] gets horizontal and vertical scrolling applied to
+    /// its uniforms, but you can disable that by calling this method. This only
+    /// applies to scrolling from its direct parent [`View`].
+    ///
+    /// This always creates a new [`DrawCall`]; no batching ever happens when using
+    /// sticky scrolling.
+    ///
+    /// TODO(JP): The fact that this only applies to the direct parent [`View`]
+    /// makes it so you can't just arbitrarily wrap [`DrawCall`]s inside [`View`]s,
+    /// which is somewhat unexpected. It might be better to have this apply to
+    /// the nearest `zaplib_components::ScrollView` instead?
+    ///
+    /// TODO(JP): Do we need to track as fields on [`DrawCall`]? The same behavior
+    /// can also be accomplished by overriding the [`Shader`]'s `scroll`
+    /// function, by doing `draw_scroll - draw_local_scroll`. It's not as
+    /// convenient, but then again, it might not be used very often, and it would
+    /// encourage people to do more stuff in shaders.
+    pub fn add_instances_with_scroll_sticky<T: Sized>(
+        &mut self,
+        shader: &'static Shader,
+        data: &[T],
+        horizontal: bool,
+        vertical: bool,
+    ) -> Area {
+        assert!(self.shader_group_instance_offsets.is_empty(), "Can't add instances with scroll sticky when in a shader group");
+
+        self.add_instances_internal(
+            shader,
+            data,
+            DrawCallProps { scroll_sticky_horizontal: horizontal, scroll_sticky_vertical: vertical, ..Default::default() },
+        )
+    }
+
+    /// Start a "shader group", which is a group of [`Shader`]s that will always be drawn in
+    /// the same order.
+    ///
+    /// When drawing the same "shader group" multiple times in a row, the existing [`DrawCall`]s
+    /// will be reused (batched) instead of new ones being created.
+    ///
+    /// For example, calling `cx.begin_shader_group(&[&FOO_SHADER, &BAR_SHADER]);` will guarantee
+    /// that for that shader group exactly two [`DrawCall`]s will be created, with `FOO_SHADER`
+    /// being always drawn first.
+    pub fn begin_shader_group(&mut self, shaders_ordered: &[&'static Shader]) {
+        assert!(self.in_redraw_cycle, "Must be in redraw cycle to call begin_shader_group");
+        assert!(self.shader_group_instance_offsets.is_empty(), "Nested shader groups are not supported (yet)");
+
+        let shader_ids: Vec<usize> = shaders_ordered.iter().map(|&shader| self.get_shader_id(shader)).collect();
+
+        // Make sure shaders are unique.
+        debug_assert!(
+            {
+                let mut unique_shader_ids = shader_ids.clone();
+                unique_shader_ids.sort_unstable();
+                unique_shader_ids.dedup();
+                shader_ids.len() == unique_shader_ids.len()
+            },
+            "Can't use shader more than once in shader group"
+        );
+
+        let shader_group_size = shaders_ordered.len();
+        let current_view_id = *self.view_stack.last().unwrap();
+        let cxview = &self.views[current_view_id];
+
+        // We have to hold the following invariant: if shader_group_instance_offsets is not empty, then the last
+        // set of `DrawCall`s in the current `CxView` have to match exactly the `shaders_ordered`. If that
+        // invariant already holds (e.g. if the exact same shader group was previously used) then we can skip
+        // creating new `DrawCall`s.
+        if self.debug_flags.disable_draw_call_batching
+            || cxview.draw_calls_len < shader_group_size
+            || shader_ids.iter().enumerate().any(|(index, &shader_id)| {
+                let dc = &cxview.draw_calls[cxview.draw_calls_len - shader_group_size + index];
+                dc.shader_id != shader_id || dc.sub_view_id != 0
+            })
+        {
+            for shader_id in shader_ids {
+                self.create_draw_call(shader_id, DrawCallProps::default());
+            }
+        }
+
+        // Since `shader_group_instance_offsets` is empty (see assertion above) we can just extend with an iterator.
+        let cxview = &self.views[current_view_id];
+        self.shader_group_instance_offsets.extend(
+            (cxview.draw_calls_len - shader_group_size..cxview.draw_calls_len)
+                .map(|index| cxview.draw_calls[index].instances.len()),
+        );
+    }
+
+    /// End a "shader group". See [`Cx::begin_shader_group`].
+    pub fn end_shader_group(&mut self) {
+        assert!(!self.shader_group_instance_offsets.is_empty(), "Call begin_shader_group before end_shader_group");
+        self.shader_group_instance_offsets.clear();
     }
 
     /// Sets the horizontal scroll position for a [`View`]/[`CxView`].
@@ -497,24 +511,6 @@ impl Cx {
     }
 }
 
-/// Temporarily get the [`DrawCall::instances`] array, so you can manually add
-/// stuff to it.
-///
-/// TODO(JP): Why do we need an object like this to be exposed to the user? Why
-/// can't we just keep the [`ManyInstances::instance_area`] directly on [`Cx`],
-/// like we do with most other `begin_` and `end_` functions (e.g. for [`Turtle`],
-/// [`View`], etc)?
-#[derive(Debug)]
-pub struct ManyInstances {
-    /// Temporarily borrowed from [`DrawCall::instances`]. You can manually add
-    /// your own instance data to it.
-    pub instances: Vec<f32>,
-    /// The [`InstanceRangeArea`] from when [`Cx::begin_many_instances`] was called.
-    /// Will be updated to include the correct count when you call
-    /// [`Cx::end_many_instances`], and returned from there as well.
-    pub(crate) instance_area: InstanceRangeArea,
-}
-
 /// Hardcoded set of uniforms that are present on every [`DrawCall`].
 ///
 /// TODO(JP): Should we just use [`Vec4`]s and [`Vec2`] here instead of individual
@@ -553,6 +549,25 @@ impl DrawUniforms {
     }
 }
 
+/// Some user-defined props to initialize a [`DrawCall`] with.
+#[derive(Default)]
+pub(crate) struct DrawCallProps {
+    /// The base [`Geometry`] object that will be used for generating the initial
+    /// vertex locations for every instance, such as a rectangle or cube.
+    /// This is currently only used when specifying custom meshes.
+    pub(crate) gpu_geometry: Option<GpuGeometry>,
+    /// See [`Cx::add_instances_with_scroll_sticky`].
+    scroll_sticky_vertical: bool,
+    /// See [`Cx::add_instances_with_scroll_sticky`].
+    scroll_sticky_horizontal: bool,
+}
+impl DrawCallProps {
+    /// Whether the draw call can be batched, or if a new one should be created.
+    fn is_batchable(&self) -> bool {
+        self.gpu_geometry.is_none() && !self.scroll_sticky_horizontal && !self.scroll_sticky_vertical
+    }
+}
+
 /// This represents an actual call to the GPU, _or_ it can represent a
 /// sub-[`View`], in case [`DrawCall::sub_view_id`] is set. Note that all of this behaves
 /// completely differently if [`DrawCall::sub_view_id`] is set; all regular drawing fields
@@ -568,8 +583,8 @@ impl DrawUniforms {
 ///
 /// It is always kept in [`CxView::draw_calls`], and as said, is part of a tree
 /// structure, called the "draw tree". To print a textual representation of the
-/// draw tree, use [`View::set_view_debug`].
-#[derive(Default, Clone)]
+/// draw tree, use [`Cx::debug_flags_mut`].
+#[derive(Default)]
 pub struct DrawCall {
     /// The index of this [`DrawCall`] within its parent [`CxView::draw_calls`].
     pub(crate) draw_call_id: usize,
@@ -581,7 +596,7 @@ pub struct DrawCall {
     /// See [`DrawCall`] for a TODO on fixing this, because this is confusing!
     pub(crate) sub_view_id: usize,
     /// The actual [`Shader`] to use when drawing.
-    pub(crate) shader: Shader,
+    pub(crate) shader_id: usize,
     /// The instance buffer that will be sent directly to the GPU.
     pub(crate) instances: Vec<f32>,
     /// Buffer of user-defined uniforms (in addition to the [`draw_uniforms`
@@ -597,54 +612,21 @@ pub struct DrawCall {
     pub(crate) uniforms_dirty: bool,
     /// Hardcoded set of uniforms that are present on every [`DrawCall`].
     pub(crate) draw_uniforms: DrawUniforms,
-    /// The base [`Geometry`] object that will be used for generating the initial
-    /// vertex locations for every instance, such as a rectangle or cube.
-    ///
-    /// TODO(JP): This geometry override seems to currently never get used, but
-    /// I can see how it could be potentially useful in the future. However, if
-    /// it doesn't turn out to be useful, let's remove it.
-    pub(crate) geometry: Option<Geometry>,
-    /// See [`DrawCall::set_scroll_sticky`].
-    pub(crate) scroll_sticky_vertical: bool,
-    /// See [`DrawCall::set_scroll_sticky`].
-    pub(crate) scroll_sticky_horizontal: bool,
-    /// Gets set to [`true`] in [`Cx::begin_many_instances`], and to [`false`] in
-    /// [`Cx::end_many_instances`], and is used to guard against doing incorrect
-    /// things while in this "many instances" mode.
-    pub(crate) in_many_instances: bool,
     /// Platform-specific data for use during painting.
     pub(crate) platform: CxPlatformDrawCall,
+    pub(crate) props: DrawCallProps,
 }
 
 impl DrawCall {
-    /// By default, [`DrawCall`] gets horizontal and vertical scrolling applied to
-    /// its uniforms, but you can disable that by calling this method. This only
-    /// applies to scrolling from its direct parent [`View`].
-    ///
-    /// TODO(JP): The fact that this only applies to the direct parent [`View`]
-    /// makes it so you can't just arbitrarily wrap [`DrawCall`]s inside [`View`]s,
-    /// which is somewhat unexpected. It might be better to have this apply to
-    /// the nearest `zaplib_widget::ScrollView` instead?
-    ///
-    /// TODO(JP): Do we need to track as fields on [`DrawCall`? The same behavior
-    /// can also be accomplished by overriding the [`Shader`]'s [`scroll`
-    /// function, by doing [`draw_scroll - draw_local_scroll`]. It's not as
-    /// convenient, but then again, it might not be used very often, and it would
-    /// encourage people to do more stuff in shaders.
-    pub fn set_scroll_sticky(&mut self, horizontal: bool, vertical: bool) {
-        self.scroll_sticky_horizontal = horizontal;
-        self.scroll_sticky_vertical = vertical;
-    }
-
     /// Set the scroll uniforms in [`DrawCall::draw_uniforms`], as computed when
     /// walking the draw tree during painting.
     pub(crate) fn set_local_scroll(&mut self, scroll: Vec2, local_scroll: Vec2) {
         self.draw_uniforms.draw_scroll_x = scroll.x;
-        if !self.scroll_sticky_horizontal {
+        if !self.props.scroll_sticky_horizontal {
             self.draw_uniforms.draw_scroll_x += local_scroll.x;
         }
         self.draw_uniforms.draw_scroll_y = scroll.y;
-        if !self.scroll_sticky_vertical {
+        if !self.props.scroll_sticky_vertical {
             self.draw_uniforms.draw_scroll_y += local_scroll.y;
         }
         self.draw_uniforms.draw_local_scroll_x = local_scroll.x;
@@ -686,6 +668,8 @@ impl DrawCall {
 /// TODO(JP): Currently empty, but I can see this be potentially useful, so I left
 /// the code around. Might want to either make use of this directly, or expose it
 /// as something users can configure, or just remove altogether.
+///  - This could potentially be used for adding transformations of many instances,
+///    for example translating or rotating, similarly to ThreeJS's Group abstraction.
 #[derive(Default, Clone)]
 #[repr(C)]
 pub struct ViewUniforms {}
@@ -696,17 +680,6 @@ impl ViewUniforms {
     }
 }
 
-/// What kind of debug information should be printed?
-///
-/// Set using [`View::set_view_debug`].
-#[derive(Clone, PartialEq)]
-pub enum CxViewDebug {
-    /// Print the draw tree.
-    DrawTree,
-    /// Print the draw tree and also information on the individual instances.
-    Instances,
-}
-
 /// An actual instantiation of a [`View`]. It's a node in the draw tree with
 /// children, which can be either [`DrawCall`]s or other [`View`]s.
 ///
@@ -717,28 +690,19 @@ pub enum CxViewDebug {
 /// NestedView(usize) }`] or so.
 ///
 /// See also [`View`] and [`ViewArea`].
-#[derive(Default, Clone)]
+#[derive(Default)]
 pub struct CxView {
     /// The actual children, which are always [`DrawCall`] objects, but those can
     /// represent either actual draw calls or child [`CxView`]s (see [`DrawCall`] and
     /// [`CxView`] for more documentation).
     pub(crate) draw_calls: Vec<DrawCall>,
-    /// The [`Rect`] of the [`Turtle`] that we created in [`View::begin_view`].
+    /// The [`Rect`] of the [`CxLayoutBox`] that we created in [`View::begin_view`].
     ///
-    /// TODO(JP): We want to decouple [`Turtle`] more from [`CxView`], so we have to
+    /// TODO(JP): We want to decouple [`CxLayoutBox`] more from [`CxView`], so we have to
     /// figure out what to do with this. For [`View`]s that are actively used for
     /// scrolling and clipping, having this `rect` makes sense, but maybe not for
     /// other uses?
     pub(crate) rect: Rect,
-    /// The id of the parent we nest in, codeflow wise.
-    ///
-    /// TODO(JP): we now have a tree in two directions: [`CxView::draw_calls`]
-    /// contains the child [`CxView`]s, and [`CxView::nesting_view_id`] points to the
-    /// parent. Is there any chance that they can get out of sync? Do we really
-    /// need both? This one here seems only to be used in [`Cx::view_will_redraw`].
-    /// If we do need to keep them, we might need some more runtime assertions to make
-    /// sure they stay in sync.
-    pub(crate) nesting_view_id: usize,
     /// The [`Cx::redraw_id`] of the last time this [`CxView`] was drawn. Can be used
     /// to see if an [`ViewArea`] pointer is still valid.
     ///
@@ -768,10 +732,10 @@ pub struct CxView {
     /// The scroll position that gets snapped to actual pixel values (taking into account
     /// the device pixel ratio; called `dpi_factor` internally).
     pub(crate) snapped_scroll: Vec2,
-    /// Platform-specific fields.
+
+    /// Platform-specific fields. Currently only used on Windows.
+    #[cfg_attr(not(target_os = "windows"), allow(dead_code))]
     pub(crate) platform: CxPlatformView,
-    /// Whether to print debug information. Set using [`View::set_view_debug`].
-    pub(crate) debug: Option<CxViewDebug>,
 }
 
 impl CxView {
@@ -787,18 +751,4 @@ impl CxView {
 
         (Vec2 { x: min_x.max(clip.0.x), y: min_y.max(clip.0.y) }, Vec2 { x: max_x.min(clip.1.x), y: max_y.min(clip.1.y) })
     }
-
-    /// Find the last [`DrawCall`] within [`CxView::draw_calls`] that matches
-    /// the given [`Shader`]. Returns an index.
-    fn find_appendable_drawcall(&mut self, shader: Shader) -> Option<usize> {
-        if self.draw_calls_len > 0 {
-            for i in (0..self.draw_calls_len).rev() {
-                let dc = &mut self.draw_calls[i];
-                if dc.sub_view_id == 0 && dc.shader == shader {
-                    return Some(i);
-                }
-            }
-        }
-        None
-    }
 }
