@@ -1,6 +1,6 @@
 //! Performance profiling.
 
-use crate::cx::*;
+use crate::*;
 
 impl Cx {
     pub fn profile_start(&mut self, id: u64) {
